import json
import argparse
from concurrent.futures import as_completed, ThreadPoolExecutor
from sliderule import sliderule, earthdata

# -------------------------------------------
# command line arguments
# -------------------------------------------

# create argument parser
parser = argparse.ArgumentParser(description="""ATL24""")
parser.add_argument('--domain',         type=str,               default="slideruleearth.io")
parser.add_argument('--organization',   type=str,               default="developers")
parser.add_argument('--granule',        type=str,               default=None) # "ATL13_20250302152414_11692601_006_01.h5"
parser.add_argument('--input_file',     type=str,               default="/data/ATL13/atl13_granules.txt")
parser.add_argument('--mapping_file',   type=str,               default="/data/ATL13/atl13_mappings.json")
parser.add_argument('--concurrency',    type=int,               default=8)
parser.add_argument('--query',          action='store_true',    default=False)  # query CMR to build list of filenames
parser.add_argument('--verbose',        action='store_true',    default=False)
parser.add_argument('--dryrun',         action='store_true',    default=False)
args,_ = parser.parse_known_args()

# initialize organization
if args.organization == "None":
    args.organization = None

# initialize client
if not args.dryrun:
    sliderule.init(args.domain, organization=args.organization, verbose=args.verbose)

# -------------------------------------------
# globals
# -------------------------------------------

atl13_id = 0
atl13_refids = {} # refid => [granule_id, ...]
atl13_granules = {} # granule_id => granule

# -------------------------------------------
# remote procedure to execute on sliderule
# -------------------------------------------

def remote_procedure(granule):
    return f"""
local json = require("json")

local asset = core.getbyname("icesat2-atl13")
local h5obj = h5.file(asset, "{granule}")

local column_gt1l = h5obj:readp("gt1l/atl13refid")
local column_gt1r = h5obj:readp("gt1r/atl13refid")
local column_gt2l = h5obj:readp("gt2l/atl13refid")
local column_gt2r = h5obj:readp("gt2r/atl13refid")
local column_gt3l = h5obj:readp("gt3l/atl13refid")
local column_gt3r = h5obj:readp("gt3r/atl13refid")

local function count(results, column)
    local values = column:unique()
    for k,_ in pairs(values) do
        results[tostring(k)] = true
    end
end

local results = {{}}
count(results, column_gt1l)
count(results, column_gt1r)
count(results, column_gt2l)
count(results, column_gt2r)
count(results, column_gt3l)
count(results, column_gt3r)

return json.encode(results)
"""

# -------------------------------------------
# worker
# -------------------------------------------

def worker (granule):
    if args.dryrun:
        return (granule, None)
    else:
        return (granule, sliderule.source("ace", remote_procedure(granule)))

# -------------------------------------------
# query CMR for ATL13 granules
# -------------------------------------------

if args.query:
    earthdata.set_max_resources(20000)
    parms = {
        "asset": "icesat2-atl13",
        "t0": '2018-01-01T00:00:00Z'
    }
    granules = earthdata.search(parms)
    print(f'Writing {len(granules)} granules to {args.input_file}')
    with open(args.input_file, "w") as file:
        for granule in granules:
            file.write(f'{granule}\n')

# -------------------------------------------
# get list of ATL13 granules to process
# -------------------------------------------

# initialize list of granules to process
granules_to_process = {}
if args.granule != None:
    granules_to_process[args.granule] = atl13_id
    atl13_granules[atl13_id] = args.granule
else:
    with open(args.input_file, "r") as file:
        for line in file.readlines():
            filename = line.strip()
            granules_to_process[filename] = atl13_id
            atl13_granules[atl13_id] = filename
            atl13_id += 1

# ---------------------------
# run multiple workers
# ---------------------------

with ThreadPoolExecutor(max_workers=args.concurrency) as executor:
    futures = [executor.submit(worker, granule) for granule in granules_to_process]
    for future in as_completed(futures):
        granule, result = future.result()
        print(f'Processed {granule}: {len(result)} reference ids')
        try:
            for refid in result:
                if refid not in atl13_refids:
                    atl13_refids[refid] = set()
                atl13_refids[refid].add(granules_to_process[granule])
        except Exception as e:
            print(f'Error parsing results for {granule}: {e}')
            print(f'Returned: {result}')

# ---------------------------
# output atl13 mappings
# ---------------------------

# convert sets to list to make json serializable
for refid in atl13_refids:
    atl13_refids[refid] = list(atl13_refids[refid])

# write out as json
with open(args.mapping_file, "w") as file:
    file.write(json.dumps({"refids": atl13_refids, "granules": atl13_granules}))



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Imports
import sliderule
from sliderule import earthdata, icesat2
import argparse
import time
import os

# Globals
ATL03_GRANULE_GRANDMESA = "ATL03_20181017222812_02950102_007_01.h5"
ATL03_GRANULE_DICKSONFJORD = "ATL03_20190314093716_11600203_007_01.h5"

# Generate Region Polygon
region = sliderule.toregion("tests/data/grandmesa.geojson")

# Command Line Arguments
parser = argparse.ArgumentParser(description="""Subset granules""")
parser.add_argument('--benchmarks',     '-b',   nargs='+', type=str,    default=[])
parser.add_argument('--domain',         '-d',   type=str,               default="slideruleearth.io")
parser.add_argument('--organization',   '-o',   type=str,               default="sliderule")
parser.add_argument('--desired_nodes',  '-n',   type=int,               default=None)
parser.add_argument('--time_to_live',   '-l',   type=int,               default=120)
parser.add_argument('--verbose',        '-v',   action='store_true',    default=False)
parser.add_argument('--loglvl',         '-j',   type=str,               default="CRITICAL")
parser.add_argument('--nocleanup',      '-u',   action='store_true',    default=False)
parser.add_argument('--no_aux',         '-x',   action='store_true',    default=False)
args,_ = parser.parse_known_args()

# Initialize Organization
if args.organization == "None":
    args.organization = None
    args.desired_nodes = None
    args.time_to_live = None

# Initialize SlideRule Client
sliderule.init(args.domain, verbose=args.verbose, loglevel=args.loglvl, organization=args.organization, desired_nodes=args.desired_nodes, time_to_live=args.time_to_live)

# Configure SlideRule Client
if args.no_aux:
    sliderule.set_processing_flags(aux=False)

# ########################################################
# Utilities
# ########################################################

# Display Results
def display_results(name, gdf, duration):
    print(f'{name} <{len(gdf)} x {len(gdf.keys())}> - {duration:.6f} secs')

# ########################################################
# Benchmarks
# ########################################################

# ------------------------------------
# Benchmark ATL06 AIO
# ------------------------------------
def atl06_aoi():
    parms = {
        "poly":             region["poly"],
        "srt":              icesat2.SRT_LAND,
    }
    return icesat2.atl06p(parms)

# ------------------------------------
# Benchmark ATL06 Ancillary
# ------------------------------------
def atl06_ancillary():
    parms = {
        "poly":             region["poly"],
        "srt":              icesat2.SRT_LAND,
        "atl03_geo_fields": ["solar_elevation"]
    }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_GRANDMESA])

# ------------------------------------
# Benchmark ATL03 Ancillary
# ------------------------------------
def atl03_ancillary():
    parms = {
        "poly":             region["poly"],
        "srt":              icesat2.SRT_LAND,
        "atl03_ph_fields":  ["ph_id_count"]
    }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_GRANDMESA])

# ------------------------------------
# Benchmark ATL06 Parquet
# ------------------------------------
def atl06_parquet():
    parms = {
        "poly":             region["poly"],
        "srt":              icesat2.SRT_LAND,
        "cnf":              icesat2.CNF_SURFACE_LOW,
        "ats":              3.0,
        "cnt":              2,
        "len":              10.0,
        "res":              10.0,
        "output":           { "path": "testfile.parquet", "format": "parquet", "open_on_complete": True } }
    gdf = icesat2.atl06p(parms, resources=[ATL03_GRANULE_GRANDMESA])
    if not args.nocleanup:
        os.remove("testfile.parquet")
    return gdf

# ------------------------------------
# Benchmark ATL03 Parquet
# ------------------------------------
def atl03_parquet():
    parms = {
        "poly":             region["poly"],
        "srt":              icesat2.SRT_LAND,
        "cnf":              icesat2.CNF_SURFACE_LOW,
        "ats":              3.0,
        "cnt":              2,
        "len":              10.0,
        "res":              10.0,
        "output":           { "path": "testfile.parquet", "format": "parquet", "open_on_complete": True } }
    gdf = icesat2.atl03sp(parms, resources=[ATL03_GRANULE_GRANDMESA])
    if not args.nocleanup:
        os.remove("testfile.parquet")
    return gdf

# ------------------------------------
# Benchmark ATL06 Sample Landsat
# ------------------------------------
def atl06_sample_landsat():
    time_start = "2021-01-01T00:00:00Z"
    time_end = "2021-02-01T23:59:59Z"
    catalog = earthdata.stac(short_name="HLS", polygon=region['poly'], time_start=time_start, time_end=time_end, as_str=True)
    parms = {
        "poly": region['poly'],
        "srt": icesat2.SRT_LAND,
        "cnf": icesat2.CNF_SURFACE_LOW,
        "ats": 20.0,
        "cnt": 10,
        "len": 40.0,
        "res": 20.0,
        "samples": {"ndvi": {"asset": "landsat-hls", "use_poi_time": True, "catalog": catalog, "bands": ["NDVI"]}} }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_GRANDMESA])

# ------------------------------------
# Benchmark ATL06 Sample (Zonal) ArcticDEM
# ------------------------------------
def atl06_sample_zonal_arcticdem():
    parms = {
        "poly": sliderule.toregion("tests/data/dicksonfjord.geojson")['poly'],
        "cnf": "atl03_high",
        "ats": 5.0,
        "cnt": 5,
        "len": 20.0,
        "res": 10.0,
        "maxi": 1,
        "samples": {"mosaic": {"asset": "arcticdem-mosaic", "radius": 10, "zonal_stats": True}} }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_DICKSONFJORD])

# ------------------------------------
# Benchmark ATL06 Sample (Nearest Neighbor) ArcticDEM
# ------------------------------------
def atl06_sample_nn_arcticdem():
    parms = {
        "poly": sliderule.toregion("tests/data/dicksonfjord.geojson")['poly'],
        "cnf": "atl03_high",
        "ats": 5.0,
        "cnt": 5,
        "len": 20.0,
        "res": 10.0,
        "maxi": 1,
        "samples": {"mosaic": {"asset": "arcticdem-mosaic"}} }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_DICKSONFJORD])

# ------------------------------------
# Benchmark ATL06 Multi-Sample (Nearest Neighbor) ArcticDEM
# ------------------------------------
def atl06_msample_nn_arcticdem():
    parms = {
        "poly": sliderule.toregion("tests/data/dicksonfjord.geojson")['poly'],
        "cnf": "atl03_high",
        "ats": 5.0,
        "cnt": 5,
        "len": 20.0,
        "res": 10.0,
        "maxi": 1,
        "samples": {"mosaic": {"asset": "arcticdem-mosaic"}} }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_DICKSONFJORD])

# ------------------------------------
# Benchmark ATL06 No Sampling ArcticDEM
# ------------------------------------
def atl06_no_sample_arcticdem():
    parms = {
        "poly": sliderule.toregion("tests/data/dicksonfjord.geojson")['poly'],
        "cnf": "atl03_high",
        "ats": 5.0,
        "cnt": 5,
        "len": 20.0,
        "res": 10.0,
        "maxi": 1 }
    return icesat2.atl06p(parms, resources=[ATL03_GRANULE_DICKSONFJORD])

# ------------------------------------
# Benchmark ATL03 Rasterized Subset
# ------------------------------------
def atl03_rasterized_subset():
    parms = {
        "poly": region['poly'],
        "region_mask": region['raster'],
        "srt": icesat2.SRT_LAND,
        "cnf": icesat2.CNF_SURFACE_LOW,
        "ats": 20.0,
        "cnt": 10,
        "len": 40.0,
        "res": 20.0 }
    return icesat2.atl03sp(parms, resources=[ATL03_GRANULE_GRANDMESA])

# ------------------------------------
# Benchmark ATL03 Polygon Subset
# ------------------------------------
def atl03_polygon_subset():
    parms = {
        "poly": region['poly'],
        "srt": icesat2.SRT_LAND,
        "cnf": icesat2.CNF_SURFACE_LOW,
        "ats": 20.0,
        "cnt": 10,
        "len": 40.0,
        "res": 20.0 }
    return icesat2.atl03sp(parms, resources=[ATL03_GRANULE_GRANDMESA])

# ########################################################
# Main
# ########################################################

if __name__ == '__main__':

    # define benchmarks
    benchmarks = {
        "atl06_aoi":                    atl06_aoi,
        "atl06_ancillary":              atl06_ancillary,
        "atl03_ancillary":              atl03_ancillary,
        "atl06_parquet":                atl06_parquet,
        "atl03_parquet":                atl03_parquet,
        "atl06_sample_landsat":         atl06_sample_landsat,
        "atl06_sample_zonal_arcticdem": atl06_sample_zonal_arcticdem,
        "atl06_sample_nn_arcticdem":    atl06_sample_nn_arcticdem,
        "atl06_msample_nn_arcticdem":   atl06_msample_nn_arcticdem,
        "atl06_no_sample_arcticdem":    atl06_no_sample_arcticdem,
        "atl03_rasterized_subset":      atl03_rasterized_subset,
        "atl03_polygon_subset":         atl03_polygon_subset,
    }

    # build list of benchmarks to run
    benchmarks_to_run = args.benchmarks
    if len(benchmarks_to_run) == 0:
        benchmarks_to_run = benchmarks.keys()

    # run benchmarks
    for benchmark in benchmarks_to_run:
        tstart = time.perf_counter()
        gdf = benchmarks[benchmark]()
        display_results(benchmark, gdf, time.perf_counter() - tstart)



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import logging
import warnings
import numpy
import tempfile
import json
import geopandas
from shapely.geometry import Polygon
from shapely.geometry.multipolygon import MultiPolygon
from datetime import datetime
from sliderule import version
from sliderule.session import Session, BASIC_TYPES, CODED_TYPE, FatalError, RetryRequest

try:
    from sklearn.cluster import KMeans
except:
    # clustering unsupported in `toregion` function
    # n_clusters restricted to a value of 1
    pass

###############################################################################
# GLOBALS
###############################################################################

DEFAULT_CRS = "EPSG:4326"
logger = logging.getLogger(__name__)
slideruleSession = Session()

###############################################################################
# APIs
###############################################################################

#
#  Initialize
#
def init (
    url=Session.PUBLIC_URL,
    verbose=False,
    loglevel=logging.INFO,
    organization=0,
    desired_nodes=None,
    time_to_live=60,
    bypass_dns=False,
    plugins=[],
    log_handler=None,
    rethrow=False ):
    '''
    Initializes the Python client for use with SlideRule, and should be called before other ICESat-2 API calls.
    This function is a wrapper for a handful of sliderule functions that would otherwise all have to be called in order to initialize the client.

    Parameters
    ----------
        url:            str
                        the IP address or hostname of the SlideRule service (slidereearth.io by default)
        verbose:        bool
                        sets up console logger as a convenience to user so all logs are printed to screen
        loglevel:       int
                        minimum severity of log message to output
        organization:   str
                        SlideRule provisioning system organization the user belongs to (see sliderule.authenticate for details)
        desired_nodes:  int
                        requested number of processing nodes in the cluster
        time_to_live:   int
                        minimum number of minutes the desired number of nodes should be present in the cluster
        bypass_dns:     bool
                        if true then the ip address for the cluster is retrieved from the provisioning system and used directly
        plugins:        list
                        names of the plugins that need to be available on the server
        log_handler:    logger
                        user provided logging handler
        rethrow:        bool
                        client rethrows exceptions to be handled by calling code

    Returns
    -------
    bool
        Status of version check

    Examples
    --------
        >>> import sliderule
        >>> sliderule.init()
    '''
    global slideruleSession
    # create new global session
    slideruleSession = Session(
        domain=url,
        verbose=verbose,
        loglevel=loglevel,
        organization=organization,
        desired_nodes=desired_nodes,
        time_to_live=time_to_live,
        bypass_dns=bypass_dns,
        log_handler=log_handler,
        rethrow=rethrow)
    # verify compatibility between client and server versions
    try:
        status = check_version(plugins=plugins)
    except Exception as e:
        logger.error(f'Version check failed: {e}')
        if rethrow:
            raise
        status = False
    return status

#
#  source
#
def source (api, parm={}, stream=False, callbacks={}, path="/source", session=None):
    '''
    Perform API call to SlideRule service

    Parameters
    ----------
        api:        str
                    name of the SlideRule endpoint
        parm:       dict
                    dictionary of request parameters
        stream:     bool
                    whether the request is a **normal** service or a **stream** service (see `De-serialization </web/rtd/user_guide/SlideRule.html#de-serialization>`_ for more details)
        callbacks:  dict
                    record type callbacks (advanced use)
        path:       str
                    path to api being requested

    Returns
    -------
    dictionary
        response data

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("slideruleearth.io")
        >>> rqst = {
        ...     "time": "NOW",
        ...     "input": "NOW",
        ...     "output": "GPS"
        ... }
        >>> rsps = sliderule.source("time", rqst)
        >>> print(rsps)
        {'time': 1300556199523.0, 'format': 'GPS'}
    '''
    session = checksession(session)
    for tolerance in [0.0001, 0.001, 0.01, 0.1, 1.0, None]:
        try:
            return session.source(api, parm=parm, stream=stream, callbacks=callbacks, path=path)
        except RetryRequest as e:
            logger.info(f'Retry requested by {api}: {e}')
            simplifypolygon(parm, tolerance)
    return None

#
#  set_url
#
def set_url (domain, session=None):
    '''
    Configure sliderule package with URL of service

    Parameters
    ----------
        urls:   str
                IP address or hostname of SlideRule service (note, there is a special case where the url is provided as a list of strings
                instead of just a string; when a list is provided, the client hardcodes the set of servers that are used to process requests
                to the exact set provided; this is used for testing and for local installations and can be ignored by most users)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("service.my-sliderule-server.org")
    '''
    session = checksession(session)
    session.service_domain = domain

#
#  set_verbose
#
def set_verbose (enable, loglevel=logging.INFO, session=None):
    '''
    Sets up a console logger to print log messages to screen

    If you want more control over the behavior of the log messages being captured, do not call this function but instead
    create and configure a Python log handler of your own and attach it to `sliderule.logger`.

    Parameters
    ----------
        enable:     bool
                    True: creates console logger if it doesn't exist, False: destroys console logger if it does exist

        loglevel:       int
                        minimum severity of log message to output

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_verbose(True, loglevel=logging.INFO)
    '''
    session = checksession(session)
    session.set_verbose(enable, loglevel)

#
# set_rqst_timeout
#
def set_rqst_timeout (timeout, session=None):
    '''
    Sets the TCP/IP connection and reading timeouts for future requests made to sliderule servers.
    Setting it lower means the client will failover more quickly, but may generate false positives if a processing request stalls or takes a long time returning data.
    Setting it higher means the client will wait longer before designating it a failed request which in the presence of a persistent failure means it will take longer for the client to remove the node from its available servers list.

    Parameters
    ----------
        timeout:    tuple
                    (<connection timeout in seconds>, <read timeout in seconds>)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_rqst_timeout((10, 60))
    '''
    session = checksession(session)
    if type(timeout) == tuple:
        session.rqst_timeout = timeout
    else:
        raise FatalError('timeout must be a tuple (<connection timeout>, <read timeout>)')

#
# set_processing_flags
#
def set_processing_flags (aux=True, session=None):
    '''
    Sets flags used when processing the record definitions

    Parameters
    ----------
        aux:    bool
                decode auxiliary fields

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_processing_flags(aux=False)
    '''
    session = checksession(session)
    if type(aux) == bool:
        session.decode_aux = aux
    else:
        raise FatalError('aux must be a boolean')


#
# update_available_servers
#
def update_available_servers (desired_nodes=None, time_to_live=None, session=None):
    '''
    Manages the number of servers in the cluster.
    If the desired_nodes parameter is set, then a request is made to change the number of servers in the cluster to the number specified.
    In all cases, the number of nodes currently running in the cluster are returned - even if desired_nodes is set;
    subsequent calls to this function is needed to check when the current number of nodes matches the desired_nodes.

    Parameters
    ----------
        desired_nodes:  int
                        the desired number of nodes in the cluster
        time_to_live:   int
                        number of minutes for the desired nodes to run

    Returns
    -------
    int
        number of nodes currently in the cluster
    int
        number of nodes available for work in the cluster

    Examples
    --------
        >>> import sliderule
        >>> num_servers, max_workers = sliderule.update_available_servers(10)
    '''
    session = checksession(session)
    return session.update_available_servers(desired_nodes, time_to_live)

#
# scaleout
#
def scaleout(desired_nodes, time_to_live, bypass_dns, session=None):
    '''
    Scale the cluster and wait for cluster to reach desired state

    Parameters
    ----------
        desired_nodes:  int
                        the desired number of nodes in the cluster
        time_to_live:   int
                        number of minutes for the desired nodes to run
        bypass_dns:     bool
                        the cluster ip address is retrieved from the provisioning system and used directly

    Examples
    --------
        >>> import sliderule
        >>> sliderule.scaleout(4, 300, False)
    '''
    session = checksession(session)
    return session.scaleout(desired_nodes, time_to_live, bypass_dns)

#
# authenticate
#
def authenticate (ps_organization, ps_username=None, ps_password=None, github_token=None, session=None):
    '''
    Authenticate to SlideRule Provisioning System
    The username and password can be provided the following way in order of priority:
    (1) The passed in arguments `github_token` or `ps_username` and `ps_password`;
    (2) The O.S. environment variables `PS_GITHUB_TOKEN` or `PS_USERNAME` and `PS_PASSWORD`;
    (3) The `ps.<url>` entry in the .netrc file in your home directory

    Parameters
    ----------
        ps_organization:    str
                            name of the SlideRule organization the user belongs to

        ps_username:        str
                            SlideRule provisioning system account name

        ps_password:        str
                            SlideRule provisioning system account password

        github_token:       str
                            GitHub access token (minimum scope/permissions require)

    Returns
    -------
    status
        True of successful, False if unsuccessful

    Examples
    --------
        >>> import sliderule
        >>> sliderule.authenticate("myorg")
        True
    '''
    session = checksession(session)
    return session.authenticate(ps_organization, ps_username, ps_password, github_token)

#
# gps2utc
#
def gps2utc (gps_time, as_str=True, session=None):
    '''
    Convert a GPS based time returned from SlideRule into a UTC time.

    Parameters
    ----------
        gps_time:   float
                    number of seconds since GPS epoch (January 6, 1980)
        as_str:     bool
                    if True, returns the time as a string; if False, returns the time as datatime object

    Returns
    -------
    datetime
        UTC time (i.e. GMT, or Zulu time)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.gps2utc(1235331234)
        '2019-02-27 19:34:03'
    '''
    session = checksession(session)
    rsps = session.source("time", {"time": int(gps_time * 1000), "input": "GPS", "output": "DATE"})
    if as_str:
        return rsps["time"]
    else:
        return datetime.strptime(rsps["time"], '%Y-%m-%dT%H:%M:%SZ')

#
# get_version
#
def get_version (session=None):
    '''
    Get the version information for the running servers and Python client

    Returns
    -------
    dict
        dictionary of version information
    '''
    session = checksession(session)
    rsps = session.source("version", {})
    if rsps != None:
        rsps["client"] = {"version": version.full_version, "organization": session.service_org}
    return rsps

#
# check_version
#
def check_version (plugins=[], session=None):
    '''
    Check that the version of the client matches the version of the server and any additionally requested plugins

    Parameters
    ----------
    plugins:    list
                list of package names (as strings) to check the version on

    Returns
    -------
    bool
        True (always; kept for backward compatibility)
    '''
    session = checksession(session)
    info = get_version(session=session)

    # check response from server
    if info == None:
        raise FatalError(f'Failed to get response from server at {session.service_org}.{session.service_domain}')

    # populate version info
    versions = {}
    for entity in ['server', 'client']:
        s = info[entity]['version'][1:].split('.')
        versions[entity] = (int(s[0]), int(s[1]), int(s[2]))

    # check major version mismatches
    if versions['server'][0] != versions['client'][0]:
        raise FatalError(f'Client {info["client"]["version"]} is incompatible with the server {info["server"]["version"]}')

    # check plugins
    for plugin in plugins:
        if versions[plugin][0] != versions['client'][0]:
            raise FatalError(f'Client {info["client"]["version"]} is incompatible with the {plugin} plugin {info[plugin]["version"]}')

    # check minor version mismatches
    if versions['server'][1] > versions['client'][1]:
        logger.warning(f'Warning, this environment is using an outdated client ({info["client"]["version"]}). The code will run but some functionality supported by the server ({info["server"]["version"]}) may not be available.')

    # return boolean for backward compatibility
    return True

#
# Run: API Request to SlideRule
#
def run(api, parms, aoi=None, resources=None, session=None):
    '''
    Execute the requested endpoint and return the results as a GeoDataFrame

    Parameters
    ----------
        api:        str
                    endpoint to run
        parms:      dict
                    parameter dictionary
        aoi:        dict
                    area of interest, passed to `sliderule.toregion()` function and polygon supplied in request
        resources:  list
                    list of resource names as strings

    Returns
    -------
    GeoDataFrame
        result of executing the request endpoint
    '''
    session = checksession(session)

    # add region
    if aoi != None:
        region = toregion(aoi)
        parms["poly"] = region["poly"]

    # add resources
    if type(resources) == list:
        parms["resources"] = resources

    # manage output for convenience
    delete_tempfile = False
    if "output" not in parms:
        delete_tempfile = True
        parms["output"] = {
            "path": tempfile.mktemp(),
            "format": "geoparquet",
            "open_on_complete": True
        }
    elif ("path" not in parms["output"]) and ("asset" not in parms["output"]):
        delete_tempfile = True
        parms["output"]["path"] = tempfile.mktemp()
        parms["output"]["open_on_complete"] = True
        if "format" not in parms["output"]:
            parms["output"]["format"] = "geoparquet"

    # make request
    rsps = source(api, {"parms": parms}, stream=True, session=session)

    # build geodataframe
    gdf = procoutputfile(parms, rsps)

    # delete tempfile
    if delete_tempfile and os.path.exists(parms["output"]["path"]):
        os.remove(parms["output"]["path"])

    # return geodataframe
    return gdf

#
# Format Region Specification
#
def toregion(source, tolerance=0.0, cellsize=0.01, n_clusters=1):
    '''
    Convert a GeoJSON/Shapefile/GeoDataFrame/list representation of a set of geospatial regions into a list of lat,lon coordinates and raster image recognized by SlideRule

    Parameters
    ----------
        source:     str
                    file name of GeoJSON formatted regions of interest, file **must** have name with the .geojson suffix
                    file name of ESRI Shapefile formatted regions of interest, file **must** have name with the .shp suffix
                    GeoDataFrame of region of interest
                    list of longitude,latitude pairs forming a polygon (e.g. [lat1, lon1, lat2, lon2, lat3, lon3, lat1, lon1])
                    list of longitude,latitude pairs forming a bounding box (e.g. [lat1, lon1, lat2, lon2])
        tolerance:  float
                    tolerance used to simplify complex shapes so that the number of points is less than the limit (a tolerance of 0.001 typically works for most complex shapes)
        cellsize:   float
                    size of pixel in degrees used to create the raster image of the polygon
        n_clusters: int
                    number of clusters of polygons to create when breaking up the request to CMR

    Returns
    -------
    dict
        a list of longitudes and latitudes containing the region of interest that can be used for the **poly** and **raster** parameters in a processing request to SlideRule.

        region = {

            "gdf": <GeoDataFrame of region>

            "poly": [{"lat": <lat1>, "lon": <lon1> }, ...],

            "raster": {"data": <geojson file as string>,

            "clusters": [[{"lat": <lat1>, "lon": <lon1>}, ...], [{"lat": <lat1>, "lon": <lon1>}, ...]] }

    Examples
    --------
    >>> import sliderule, json
    >>> region = sliderule.toregion("tests/data/grandmesa.geojson")
    >>> print(json.dumps(region["poly"], indent=4))
    [
        {
            "lon": -108.20772968780051,
            "lat": 38.8232055291981
        },
        {
            "lon": -108.07460164311031,
            "lat": 38.8475137825863
        },
        {
            "lon": -107.72839858755752,
            "lat": 39.01510930230633
        },
        {
            "lon": -107.78724142490994,
            "lat": 39.195630349659986
        },
        {
            "lon": -108.17287000970857,
            "lat": 39.15920066396116
        },
        {
            "lon": -108.31168256553767,
            "lat": 39.13757646212944
        },
        {
            "lon": -108.34115668325224,
            "lat": 39.03758987613325
        },
        {
            "lon": -108.2878686387796,
            "lat": 38.89051431295789
        },
        {
            "lon": -108.20772968780051,
            "lat": 38.8232055291981
        }
    ]
    '''
    # GeoDataFrame
    if isinstance(source, geopandas.GeoDataFrame):
        gdf = source
        datafile = gdf.to_json()

    # Polygon
    elif isinstance(source, Polygon):
        gdf = geopandas.GeoDataFrame(geometry=[source], crs=DEFAULT_CRS)
        datafile = gdf.to_json()

    # bounding box as [lon lat lon lat]
    elif isinstance(source, list) and (len(source) >= 4) and (len(source) % 2 == 0):
        if len(source) == 4: # bounding box
            lons = [source[0], source[2], source[2], source[0], source[0]]
            lats = [source[1], source[1], source[3], source[3], source[1]]
        elif len(source) > 4: # polygon list
            lons = [source[i] for i in range(1,len(source),2)]
            lats = [source[i] for i in range(0,len(source),2)]
        p = Polygon([point for point in zip(lons, lats)])
        gdf = geopandas.GeoDataFrame(geometry=[p], crs=DEFAULT_CRS)
        datafile = gdf.to_json()

    # List of lat/lon dictionaries
    elif isinstance(source, list) and (len(source) >= 4) and isinstance(source[0], dict):
        p = Polygon([(c["lon"], c["lat"]) for c in source])
        gdf = geopandas.GeoDataFrame(geometry=[p], crs=DEFAULT_CRS)
        datafile = gdf.to_json()

    # Shapefile
    elif isinstance(source, str) and (source.find(".shp") > 1):
        gdf = geopandas.read_file(source)
        datafile = gdf.to_json()

    # GeoJSON file
    elif isinstance(source, str) and (source.find(".geojson") > 1):
        gdf = geopandas.read_file(source)
        with open(source, mode='rt') as file:
            datafile = file.read()

    # GeoJSON dictionary
    elif isinstance(source, dict) and ("features" in source):
        gdf = geopandas.GeoDataFrame.from_features(source["features"])
        datafile = gdf.to_json()

    # GeoJSON string
    elif isinstance(source, str) and (source.find("FeatureCollection") > 0):
        geojson_dict = json.loads(source)
        gdf = geopandas.GeoDataFrame.from_features(geojson_dict["features"])
        datafile = gdf.to_json()

    # Unsupported
    else:
        raise FatalError("incorrect filetype: please use a .geojson, .shp, or a geodataframe")

    # If user provided raster we don't have gdf, geopandas cannot easily convert it
    polygon = clusters = None
    if gdf is not None:
        # simplify polygon
        if(tolerance > 0.0):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                gdf = gdf.buffer(tolerance)
                gdf = gdf.simplify(tolerance)

        # generate polygon
        polygon = gdf2poly(gdf)

        # generate clusters
        clusters = []
        if n_clusters > 1:
            # pull out centroids of each geometry object
            if "CenLon" in gdf and "CenLat" in gdf:
                X = numpy.column_stack((gdf["CenLon"], gdf["CenLat"]))
            else:
                s = gdf.centroid
                X = numpy.column_stack((s.x, s.y))
            # run k means clustering algorithm against polygons in gdf
            kmeans = KMeans(n_clusters=n_clusters, init='k-means++', random_state=5, max_iter=400)
            y_kmeans = kmeans.fit_predict(X)
            k = geopandas.pd.DataFrame(y_kmeans, columns=['cluster'])
            gdf = gdf.join(k)
            # build polygon for each cluster
            for n in range(n_clusters):
                c_gdf = gdf[gdf["cluster"] == n]
                c_poly = gdf2poly(c_gdf)
                clusters.append(c_poly)

    # return region
    return {
        "gdf": gdf,
        "poly": polygon, # convex hull of polygons
        "clusters": clusters, # list of polygon clusters for cmr request
        "raster": {
            "geojson": datafile, # geojson file
            "length": len(datafile), # geojson file length
            "cellsize": cellsize  # units are in crs/projection
        }
    }

###############################################################################
# INTERNAL APIs
###############################################################################

#
# Get Record Field Definition
#
def getdefinition (rectype, fieldname, session=None):
    session = checksession(session)
    return session.get_definition(rectype, fieldname)
#
# GeoDataFrame to Polygon
#
def gdf2poly(gdf):

    # pull out coordinates
    hull = gdf.unary_union.convex_hull
    polygon = [{"lon": coord[0], "lat": coord[1]} for coord in list(hull.exterior.coords)]

    # determine winding of polygon #
    #              (x2               -    x1)             *    (y2               +    y1)
    wind = sum([(polygon[i+1]["lon"] - polygon[i]["lon"]) * (polygon[i+1]["lat"] + polygon[i]["lat"]) for i in range(len(polygon) - 1)])
    if wind > 0:
        # reverse direction (make counter-clockwise) #
        ccw_poly = []
        for i in range(len(polygon), 0, -1):
            ccw_poly.append(polygon[i - 1])
        # replace region with counter-clockwise version #
        polygon = ccw_poly

    # return polygon
    return polygon

#
# Create Empty GeoDataFrame
#
def emptyframe(**kwargs):
    # set default keyword arguments
    kwargs.setdefault('crs',DEFAULT_CRS)
    return geopandas.GeoDataFrame(geometry=geopandas.points_from_xy([], [], []), crs=kwargs['crs'])

#
# Get Values from Raw Buffer
#
def getvalues(data, dtype, size, num_elements=0):
    """
    data:   tuple of bytes
    dtype:  element of codedtype
    size:   bytes in data
    """
    raw = bytes(data)
    datatype = BASIC_TYPES[CODED_TYPE[dtype]]["nptype"]
    if num_elements == 0: # dynamically determine number of elements
        num_elements = int(size / numpy.dtype(datatype).itemsize)
    slicesize = num_elements * numpy.dtype(datatype).itemsize # truncates partial bytes
    values = numpy.frombuffer(raw[:slicesize], dtype=datatype, count=num_elements)
    return values

#
# Process Output File
#
def procoutputfile(parm, rsps):
    '''
    parm: request parameters
    rsps: response
    '''
    output = parm["output"]

    # Get file path to read from
    if "path" in output:
        path = output["path"]
    else:
        path = None

    # Check if it is a remote path
    for rsp in rsps:
        if 'arrowrec.remote' == rsp['__rectype']:
            path = rsp['url']
            break

    # Handle local files
    local_file = path # default to just returning the name of the parquet file
    if "open_on_complete" in output and output["open_on_complete"]:

        # Open the file as a DataFrame
        if (output["format"] == "geoparquet") or (output["format"] == "parquet" and "as_geo" in output and output["as_geo"]):
            local_file = geopandas.read_parquet(path) # GeoParquet
        elif output["format"] == "parquet":
            local_file = geopandas.pd.read_parquet(path) # Parquet
        elif output["format"] == "feather":
            local_file = geopandas.pd.read_feather(path) # Feather
        elif output["format"] == "csv":
            local_file = geopandas.pd.read_csv(path) # CSV
        else:
            raise FatalError('unsupported output format: %s' % (output["format"]))

        # Read metadata from file
        try:
            # Imports needed just to read metadata
            import pyarrow.parquet as pq
            import ctypes
            import json
            # pull out metadata
            metadata = pq.read_metadata(path)
            for key in metadata.metadata:
                if key in [b'ARROW:schema', b'pandas', b'geo']:
                    continue
                metadata_str = ctypes.create_string_buffer(metadata.metadata[key]).value.decode('ascii')
                local_file.attrs[key.decode('ascii')] = json.loads(metadata_str)
        except Exception as e:
            # could fail for a multitude of reasons; just log and move on
            logger.debug(f'Failed to read metadata from {path}: {e}')

    # Return back to caller either path or opened dataframe
    return local_file

#
# Dictionary to GeoDataFrame
#
def todataframe(columns, time_key="time", lon_key="longitude", lat_key="latitude", height_key=None, **kwargs):

    # Set Default Keyword Arguments
    kwargs.setdefault('index_key','time')
    kwargs.setdefault('crs',DEFAULT_CRS)

    # Check Empty Columns
    if len(columns) <= 0:
        return emptyframe(**kwargs)

    # Generate Time Column
    columns['time'] = columns[time_key].astype('datetime64[ns]')

    # Generate Geometry Column
    # 3D point geometry
    # This enables 3D CRS transformations using the to_crs() method
    if height_key == None:
        geometry = geopandas.points_from_xy(columns[lon_key], columns[lat_key])
    else:
        geometry = geopandas.points_from_xy(columns[lon_key], columns[lat_key], columns[height_key])
    del columns[lon_key]
    del columns[lat_key]

    # Create Pandas DataFrame object
    if type(columns) == dict:
        df = geopandas.pd.DataFrame(columns)
    else:
        df = columns

    # Build GeoDataFrame (default geometry is crs=SLIDERULE_EPSG)
    gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=kwargs['crs'])

    # Set index (default is Timestamp), can add `verify_integrity=True` to check for duplicates
    # Can do this during DataFrame creation, but this allows input argument for desired column
    gdf.set_index(kwargs['index_key'], inplace=True)

    # Sort values for reproducible output despite async processing
    gdf.sort_index(inplace=True)

    # Return GeoDataFrame
    return gdf

#
# Simplify Polygon
#
def simplifypolygon(parm, tolerance):
    if "parms" not in parm:
        return

    if "cmr" in parm["parms"]:
        polygon = parm["parms"]["cmr"]["polygon"]
    elif "poly" in parm["parms"]:
        polygon = parm["parms"]["poly"]
    else:
        return

    raw_multi_polygon = [[(tuple([(c['lon'], c['lat']) for c in polygon]), [])]]
    shape = MultiPolygon(*raw_multi_polygon)
    buffered_shape = shape.buffer(tolerance)
    simplified_shape = buffered_shape.simplify(tolerance)
    simplified_coords = list(simplified_shape.exterior.coords)

    simplified_polygon = []
    for coord in simplified_coords:
        point = {"lon": coord[0], "lat": coord[1]}
        simplified_polygon.insert(0, point)

    if "cmr" not in parm["parms"]:
        parm["parms"]["cmr"] = {}
    parm["parms"]["cmr"]["polygon"] = simplified_polygon

    logger.warning('Using simplified polygon (for CMR request only!), {} points using tolerance of {}'.format(len(simplified_coords), tolerance))

#
# Set Session
#
def checksession(session):
    global slideruleSession
    return session if session != None else slideruleSession



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from sliderule import sliderule, icesat2
import argparse
import time
import threading
import queue
import logging
import sys
import glob
import boto3

# Command Line Arguments
parser = argparse.ArgumentParser(description="""ATL24""")
parser.add_argument('--domain',             type=str,               default="slideruleearth.io")
parser.add_argument('--organization',       type=str,               default="bathy")
parser.add_argument('--desired_nodes',      type=int,               default=None)
parser.add_argument('--time_to_live',       type=int,               default=600) # 10 hours
parser.add_argument('--log_file',           type=str,               default="atl24.log")
parser.add_argument('--concurrency',        type=int,               default=50)
parser.add_argument('--startup_separation', type=int,               default=2) # seconds
parser.add_argument('--exclude_empty',      action='store_true',    default=False)
parser.add_argument('--input_file',         type=str,               default="testsets/bluetopo_granules.txt")
parser.add_argument('--input_files',        type=str,               default=None) # glob support
parser.add_argument('--url',                type=str,               default="s3://sliderule-public")
parser.add_argument('--retries',            type=int,               default=3)
parser.add_argument('--report_only',        action='store_true',    default=False)
args,_ = parser.parse_known_args()

# Initialize Constants
RELEASE = "001"
VERSION = "01"

# Initialize Organization
if args.organization == "None":
    args.organization = None
    args.desired_nodes = None
    args.time_to_live = None

# Setup Log File
LOG_FORMAT = '%(created)f %(levelname)-5s [%(filename)s:%(lineno)5d] %(message)s'
log        = logging.getLogger(__name__)
logfmt     = logging.Formatter(LOG_FORMAT)
logfile    = logging.FileHandler(args.log_file)
logfile.setFormatter(logfmt)
log.addHandler(logfile)
log.setLevel(logging.INFO)
logfile.setLevel(logging.INFO)

# Initialize Concurrent Requests
rqst_q = queue.Queue()
if args.concurrency != None:
    concurrent_rqsts = args.concurrency
elif args.desired_nodes != None:
    concurrent_rqsts = args.desired_nodes
else:
    concurrent_rqsts = 1

# Creat S3 Client
s3_client = boto3.client("s3")

#
# List H5 Granules
#
def list_h5_granules(url, input_files):

    # get bucket and subfolder from url
    path = url.split("s3://")[-1]
    bucket = path.split("/")[0]
    subfolder = '/'.join(path.split("/")[1:])

    # read granules
    granules = []
    is_truncated = True
    continuation_token = None
    while is_truncated:
        # make request
        if continuation_token:
            response = s3_client.list_objects_v2(Bucket=bucket, Prefix=subfolder, ContinuationToken=continuation_token)
        else:
            response = s3_client.list_objects_v2(Bucket=bucket, Prefix=subfolder)
        # display status
        print("#", end='')
        sys.stdout.flush()
        # parse contents
        if 'Contents' in response:
            for obj in response['Contents']:
                resource = obj['Key'].split("/")[-1]
                if resource.startswith("ATL24") and \
                  (resource.endswith("_" + RELEASE + "_" + VERSION + ".h5") or \
                   (not args.exclude_empty and resource.endswith(".empty"))):
                    granules.append(resource[:36].replace("ATL24", "ATL03") + ".h5")
        # check if more data is available
        is_truncated = response['IsTruncated']
        continuation_token = response.get('NextContinuationToken')
    print('') # new line

    # open and read input file
    rqsts = []
    already_processed = 0
    for input_file in input_files:
        with open(input_file, "r") as file:
            for line in file.readlines():
                granule = line.strip()
                if granule not in granules:
                    rqsts.append(granule)
                else:
                    already_processed += 1

    # Display Parameters
    print(f'Existing Granules:           {len(granules)}')
    print(f'Granules Already Processed   {already_processed}')
    print(f'Granules Left to Process:    {len(rqsts)}')

    # return list of granules to process
    return rqsts

#
# Process Request
#
def process_request(worker_id, granule):

    # Start Processing
    log.info(f'<{worker_id}> processing granule {granule} ...')

    # Build Output File Path
    output_filename = granule.replace("ATL03", "ATL24")[:-3] + "_" + RELEASE + "_" + VERSION + ".h5"

    # Set Parameters
    parms = {
        "asset": "icesat2",
        "resource": granule,
        "atl09_asset": "icesat2",
        "srt": icesat2.SRT_DYNAMIC,
        "cnf": "atl03_not_considered",
        "pass_invalid": True,
        "use_bathy_mask": True,
        "timeout": 600, # 10 minutes
        "classifiers": ["qtrees", "coastnet", "medianfilter", "cshelph", "bathypathfinder", "openoceanspp", "ensemble"],
        "output": {
            "asset": "sliderule-stage",
            "path": output_filename,
            "format": "h5",
            "open_on_complete": False,
            "with_checksum": False,
        }
    }

    # Make Request
    outfile = icesat2.atl24g(parms)
    log.info(f'<{worker_id}> finished granule {outfile}')

    # Parse Outfile Name
    path = outfile.split("s3://")[-1]
    bucket = path.split("/")[0]
    object_key = '/'.join(path.split("/")[1:])

    # Check Empty
    if outfile.endswith(".empty"):
        response = s3_client.get_object(Bucket=bucket, Key=object_key)
        content = response["Body"].read().decode("utf-8")
        return f'Empty - {content}'

    # Check Size
    try:
        response = s3_client.head_object(Bucket=bucket, Key=object_key)
        size = response["ContentLength"]
        if size <= 0:
            return f'Missing - size = {size}'
    except Exception as e:
        return "Missing"

    # Success
    return "Valid"

#
# Worker
#
def worker(worker_id):

    # While Queue Not Empty
    complete = False
    while not complete:

        # Check Server Availability
        while True:
            available_servers, _ = sliderule.update_available_servers()
            if (worker_id + 1) <= available_servers:
                break
            else:
                log.info(f'<{worker_id}> waiting for servers to become available')
                time.sleep(30)

        # Get Request
        try:
            granule = rqst_q.get(block=False)
        except Exception as e:
            # Handle No More Requests
            if rqst_q.empty():
                log.info(f'<{worker_id}> no more requests {e}')
                complete = True
            else:
                log.info(f'<{worker_id}> exception: {e}')
            time.sleep(5) # prevents a spin
            continue

        # Process Request
        valid = False
        status = "Unknown"
        attempts = args.retries
        while not valid and attempts > 0:
            attempts -= 1
            try:
                status = process_request(worker_id, granule)
                if status != "Missing":
                    valid = True
                else:
                    time.sleep(30) # slow it down
            except Exception as e:
                log.critical(f'failure on {granule}: {e}')
                time.sleep(30) # slow it down
                status = f'Error - {e}'

        # Display Status
        print(f'{granule}: {status}')

# Get List of Input Files
if args.input_files == None:
    list_of_input_files = [args.input_file]
else:
    list_of_input_files = glob.glob(args.input_files)

# Generate List of Unprocessed Granules
rqsts = list_h5_granules(args.url, list_of_input_files)

# Queue Processing Requests
for granule in rqsts:
    if "#" not in granule:
        rqst_q.put(granule)

# Check for Early Exit
if args.report_only:
    sys.exit(0)

# Initialize Python Client
sliderule.init(args.domain, verbose=False, loglevel=logging.INFO, organization=args.organization, desired_nodes=args.desired_nodes, time_to_live=args.time_to_live, rethrow=True, log_handler=logfile)

# Start Workers
threads = [threading.Thread(target=worker, args=(worker_id,), daemon=True) for worker_id in range(concurrent_rqsts)]
for thread in threads:
    thread.start()
    time.sleep(args.startup_separation)
for thread in threads:
    thread.join()
log.info('all processing requests completed')



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import logging
import numpy
import sliderule
from sliderule import logger
from sliderule.session import Session

###############################################################################
# APIs
###############################################################################

#
#  Initialize
#
def init (url=Session.PUBLIC_URL, verbose=False, loglevel=logging.CRITICAL, organization=Session.PUBLIC_ORG, desired_nodes=None, time_to_live=60, bypass_dns=False):
    '''
    Initializes the Python client for use with SlideRule and should be called before other SWOT API calls.
    This function is a wrapper for the `sliderule.init(...) function </web/rtds/api_reference/sliderule.html#init>`_.

    Examples
    --------
        >>> from sliderule import swot
        >>> swot.init()
    '''
    sliderule.init(url, verbose, loglevel, organization, desired_nodes, time_to_live, bypass_dns)

#
#  L2
#
def swotl2 (parm, resource):
    '''
    Performs L2 subsetting of swaths

    Parameters
    ----------
    parms:      dict
                parameters used to configure subsetting process
    resource:   str
                SWOT NetCDF filename
    asset:      str
                data source asset

    Returns
    -------
    dict
        dictionary of requested datasets keyed by granule name and dataset name
    '''
    return swotl2p(parm, resources=[resource])

#
#  Parallel L2
#
def swotl2p(parm, callbacks={}, resources=None):
    '''
    Performs subsetting in parallel on SWOT  data and returns swaths.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure subsetting process
        asset:          str
                        data source asset
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["SWOT_L2_LR_SSH_Expert_009_009_20111121T053342_20111121T062448_DG10_01.nc", ...])

    Returns
    -------
    dict
        dictionary of requested datasets keyed by granule name and dataset name

    Examples
    --------
        >>> from sliderule import swot
        >>> swot.init()
        >>> region = [ {"lon":-105.82971551223244, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 39.81983728534918} ]
        >>> parms = { "poly": region }
        >>> resources = ["SWOT_L2_LR_SSH_Expert_009_009_20111121T053342_20111121T062448_DG10_01.nc"]
        >>> asset = "swot-sim-ecco-llc4320"
        >>> rsps = swot.swotl2p(parms, asset=asset, resources=resources)
    '''
    try:
        # Default the Asset
        rqst_parm = parm.copy()
        if "asset" not in rqst_parm:
            rqst_parm["asset"] = "swot-sim-ecco-llc4320"

        # Build GEDI Request
        rqst = {
            "resources": resources,
            "parms": rqst_parm
        }

        # Make API Processing Request
        rsps = sliderule.source('swotl2p', rqst, stream=True, callbacks=callbacks)

        # Process Responses
        results = {}
        if len(rsps) > 0:
            for rsp in rsps:
                if 'swotl2var' in rsp['__rectype']:
                    granule = rsp['granule']
                    variable = rsp['variable']
                    if granule not in results:
                        results[granule] = {}
                    if variable not in results[granule]:
                        results[granule][variable] = {}
                    data = sliderule.getvalues(rsp['data'], rsp['datatype'], len(rsp['data']))
                    num_rows = int(rsp['elements'] / rsp['width'])
                    results[granule][variable] = data.reshape(num_rows, rsp['width'])
                elif 'swotl2geo' == rsp['__rectype']:
                    granule = rsp['granule']
                    if granule not in results:
                        results[granule] = {}
                    latitudes = []
                    longitudes = []
                    for point in rsp['scan']:
                        latitudes += point['latitude'],
                        longitudes += point['longitude'],
                    results[granule]['latitude'] = numpy.array(latitudes)
                    results[granule]['longitude'] = numpy.array(longitudes)

        # Return Results
        return results

    # Handle Runtime Errors
    except RuntimeError as e:
        logger.critical(e)
        return {}



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import numpy
import geopandas
import sliderule
from sliderule import earthdata, logger
from sliderule.session import Session, BASIC_TYPES, CODED_TYPE

###############################################################################
# GLOBALS
###############################################################################

# CRS used in ICESat-2 Standard Data Products
ICESAT2_CRS = "EPSG:7912"

# icesat2 parameters
CNF_POSSIBLE_TEP = -2
CNF_NOT_CONSIDERED = -1
CNF_BACKGROUND = 0
CNF_WITHIN_10M = 1
CNF_SURFACE_LOW = 2
CNF_SURFACE_MEDIUM = 3
CNF_SURFACE_HIGH = 4
SRT_DYNAMIC = -1
SRT_LAND = 0
SRT_OCEAN = 1
SRT_SEA_ICE = 2
SRT_LAND_ICE = 3
SRT_INLAND_WATER = 4
MAX_COORDS_IN_POLYGON = 16384
GT1L = 10
GT1R = 20
GT2L = 30
GT2R = 40
GT3L = 50
GT3R = 60
STRONG_SPOTS = (1, 3, 5)
WEAK_SPOTS = (2, 4, 6)
LEFT_PAIR = 0
RIGHT_PAIR = 1
SC_BACKWARD = 0
SC_FORWARD = 1
ATL08_WATER = 0
ATL08_LAND = 1
ATL08_SNOW = 2
ATL08_ICE = 3

# phoreal percentiles
P = { '5':   0, '10':  1, '15':  2, '20':  3, '25':  4, '30':  5, '35':  6, '40':  7, '45':  8, '50': 9,
      '55': 10, '60': 11, '65': 12, '70': 13, '75': 14, '80': 15, '85': 16, '90': 17, '95': 18 }

###############################################################################
# APIs
###############################################################################

#
#  Initialize
#
def init (url=Session.PUBLIC_URL, verbose=False, max_resources=earthdata.DEFAULT_MAX_REQUESTED_RESOURCES, loglevel=logging.CRITICAL, organization=Session.PUBLIC_ORG, desired_nodes=None, time_to_live=60, bypass_dns=False, rethrow=False):
    '''
    Initializes the Python client for use with SlideRule and should be called before other ICESat-2 API calls.
    This function is a wrapper for the `sliderule.init(...) function </web/rtds/api_reference/sliderule.html#init>`_.

    Parameters
    ----------
        max_resources:  int
                        maximum number of H5 granules to process in the request

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init()
    '''
    status = sliderule.init(url, verbose=verbose, loglevel=loglevel, organization=organization, desired_nodes=desired_nodes, time_to_live=time_to_live, bypass_dns=bypass_dns, rethrow=rethrow)
    earthdata.set_max_resources(max_resources) # set maximum number of resources allowed per request
    return status

#
#  ATL06
#
def atl06 (parm, resource):
    '''
    Performs ATL06-SR processing on ATL03 data and returns geolocated elevations

    Parameters
    ----------
    parms:      dict
                parameters used to configure ATL06-SR algorithm processing (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
    resource:   str
                ATL03 HDF5 filename

    Returns
    -------
    GeoDataFrame
        geolocated elevations (see `Elevations </web/rtd/user_guide/icesat2.html#elevations>`_)
    '''
    return atl06p(parm, resources=[resource])

#
#  Parallel ATL06
#
def atl06p(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs ATL06-SR processing in parallel on ATL03 data and returns geolocated elevations.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Warnings
    --------
        It is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement
        to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL06-SR algorithm processing (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_007_01.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        geolocated elevations (see `Elevations </web/rtd/user_guide/icesat2.html#elevations>`_)

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init("slideruleearth.io", True)
        >>> parms = { "cnf": 4, "ats": 20.0, "cnt": 10, "len": 40.0, "res": 20.0 }
        >>> resources = ["ATL03_20181019065445_03150111_003_01.h5"]
        >>> atl03_asset = "atlas-local"
        >>> rsps = icesat2.atl06p(parms, asset=atl03_asset, resources=resources)
        >>> rsps
                dh_fit_dx  w_surface_window_final  ...                       time                     geometry
        0        0.000042               61.157661  ... 2018-10-19 06:54:46.104937  POINT (-63.82088 -79.00266)
        1        0.002019               61.157683  ... 2018-10-19 06:54:46.467038  POINT (-63.82591 -79.00247)
        2        0.001783               61.157678  ... 2018-10-19 06:54:46.107756  POINT (-63.82106 -79.00283)
        3        0.000969               61.157666  ... 2018-10-19 06:54:46.469867  POINT (-63.82610 -79.00264)
        4       -0.000801               61.157665  ... 2018-10-19 06:54:46.110574  POINT (-63.82124 -79.00301)
        ...           ...                     ...  ...                        ...                          ...
        622407  -0.000970               61.157666  ... 2018-10-19 07:00:29.606632  POINT (135.57522 -78.98983)
        622408   0.004620               61.157775  ... 2018-10-19 07:00:29.250312  POINT (135.57052 -78.98983)
        622409  -0.001366               61.157671  ... 2018-10-19 07:00:29.609435  POINT (135.57504 -78.98966)
        622410  -0.004041               61.157748  ... 2018-10-19 07:00:29.253123  POINT (135.57034 -78.98966)
        622411  -0.000482               61.157663  ... 2018-10-19 07:00:29.612238  POINT (135.57485 -78.98948)

        [622412 rows x 16 columns]
    '''
    # Build Request
    rqst = __build_request(parm, resources)

    # Make API Processing Request
    rsps = sliderule.source("atl06p", rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches(rsps, 'atl06rec', 'elevation', parm, keep_id, as_numpy_array, height_key)

#
#  Subsetted ATL06
#
def atl06s (parm, resource):
    '''
    Subsets ATL06 data given the polygon and time range provided and returns elevations

    Parameters
    ----------
        parms:      dict
                    parameters used to configure ATL03 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        resource:   str
                    ATL06 HDF5 filename

    Returns
    -------
    GeoDataFrame
        ATL06 elevations
    '''
    return atl06sp(parm, resources=[resource])

#
#  Parallel Subsetted ATL06
#
def atl06sp(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs ATL06 subsetting in parallel on ATL06 data and returns elevation data.  Unlike the `atl06s <#atl06s>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Warnings
    --------
        Note, it is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL03 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_007_01.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        ATL06 elevations
    '''
    # Build Request
    rqst = __build_request(parm, resources, default_asset="icesat2-atl06")

    # Make API Processing Request
    rsps = sliderule.source("atl06sp", rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches(rsps, 'atl06srec', 'elevation', parm, keep_id, as_numpy_array, height_key)

#
#  Subsetted ATL13
#
def atl13s (parm, resource):
    '''
    Subsets ATL13 data given the polygon and time range provided and returns measurements

    Parameters
    ----------
        parms:      dict
                    parameters used to configure ATL13 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        resource:   str
                    ATL13 HDF5 filename

    Returns
    -------
    GeoDataFrame
        ATL13 water measurements
    '''
    return atl13sp(parm, resources=[resource])

#
#  Parallel Subsetted ATL13
#
def atl13sp(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs ATL13 subsetting in parallel on ATL13 data and returns measurement data.  Unlike the `atl13s <#atl13s>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL13 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL13_20181019065445_03150111_007_01.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        ATL13 water measurements
    '''
    # Build Request
    rqst = __build_request(parm, resources, default_asset="icesat2-atl13")

    # Make API Processing Request
    rsps = sliderule.source("atl13sp", rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches(rsps, 'atl13srec', 'water', parm, keep_id, as_numpy_array, height_key)

#
#  Subsetted ATL03
#
def atl03s (parm, resource):
    '''
    Subsets ATL03 data given the polygon and time range provided and returns segments of photons

    Parameters
    ----------
        parms:      dict
                    parameters used to configure ATL03 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        resource:   str
                    ATL03 HDF5 filename

    Returns
    -------
    GeoDataFrame
        ATL03 extents (see `Photon Segments </web/rtd/user_guide/icesat2.html#segmented-photon-data>`_)
    '''
    return atl03sp(parm, resources=[resource])

#
#  Parallel Subsetted ATL03
#
def atl03sp(parm, callbacks={}, resources=None, keep_id=False, height_key=None):
    '''
    Performs ATL03 subsetting in parallel on ATL03 data and returns photon segment data.  Unlike the `atl03s <#atl03s>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Warnings
    --------
        Note, it is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL03 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_007_01.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        ATL03 segments (see `Photon Segments </web/rtd/user_guide/icesat2.html#photon-segments>`_)
    '''
    rqst = __build_request(parm, resources)

    # Make Request
    rsps = sliderule.source("atl03sp", rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches03(rsps, parm, keep_id, height_key)

#
#  Viewer ATL03
#
def atl03v (parm, resource):
    '''
    Subsets ATL03 data given the polygon and time range provided and returns counts of photons in segments

    Parameters
    ----------
        parms:      dict
                    parameters used to configure ATL03 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        resource:   str
                    ATL03 HDF5 filename

    Returns
    -------
    GeoDataFrame
        ATL03 extents (see `Photon Segments </web/rtd/user_guide/icesat2.html#segmented-photon-data>`_)
    '''
    return atl03vp(parm, resources=[resource])

#
#  Parallel Viewer ATL03
#
def atl03vp(parm, callbacks={}, resources=None, keep_id=False):
    '''
    Performs ATL03 subsetting in parallel on ATL03 data and returns counts of photons in segments.  Unlike the `atl03v <#atl03v>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Warnings
    --------
        Note, it is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL03 subsetting (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_007_01.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges

    Returns
    -------
    GeoDataFrame
        ATL03 segments (see `Photon Segments </web/rtd/user_guide/icesat2.html#photon-segments>`_)
    '''
    # Build Request
    rqst = __build_request(parm, resources)

    # Make Request
    rsps = sliderule.source("atl03vp", rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches03v(rsps, parm, keep_id)

#
#  ATL08
#
def atl08 (parm, resource):
    '''
    Performs ATL08-PhoREAL processing on ATL03 and ATL08 data and returns geolocated elevations

    Parameters
    ----------
    parms:      dict
                parameters used to configure ATL06-SR algorithm processing (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
    resource:   str
                ATL03 HDF5 filename

    Returns
    -------
    GeoDataFrame
        geolocated vegatation statistics
    '''
    return atl08p(parm, resources=[resource])

#
#  Parallel ATL08
#
def atl08p(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs ATL08-PhoREAL processing in parallel on ATL03 and ATL08 data and returns geolocated vegatation statistics.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Warnings
    --------
        It is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement
        to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL06-SR algorithm processing (see `Parameters </web/rtd/user_guide/icesat2.html#parameters>`_)
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_007_01.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        geolocated vegetation statistics
    '''
    # Build Request
    rqst = __build_request(parm, resources)

    # Make Request
    rsps = sliderule.source("atl08p", rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches(rsps, 'atl08rec', 'vegetation', parm, keep_id, as_numpy_array, height_key)

#
#  ATL24 Photon Viewer
#
def atl24v(parm, resource):
    '''
    Provides statistics for the number of photons within a granule
    '''
    # Default the Asset
    if "asset" not in parm:
        parm["asset"] = "icesat2"

    # Build Request
    rqst = {
        "resource": resource,
        "parms": parm
    }

    # Make Request
    rsps = sliderule.source("atl24v", rqst, stream=False)

    # Return Response
    return rsps

###############################################################################
# LOCAL FUNCTIONS
###############################################################################

#
# Calculate Laser Spot
#
def __calcspot(df):

    # Create dictionary mapping (sc_orient, track, pair) to spot
    map_spot = {(SC_BACKWARD,   1,  LEFT_PAIR):     1,
                (SC_BACKWARD,   1,  RIGHT_PAIR):    2,
                (SC_BACKWARD,   2,  LEFT_PAIR):     3,
                (SC_BACKWARD,   2,  RIGHT_PAIR):    4,
                (SC_BACKWARD,   3,  LEFT_PAIR):     5,
                (SC_BACKWARD,   3,  RIGHT_PAIR):    6,
                (SC_FORWARD,    1,  LEFT_PAIR):     6,
                (SC_FORWARD,    1,  RIGHT_PAIR):    5,
                (SC_FORWARD,    2,  LEFT_PAIR):     4,
                (SC_FORWARD,    2,  RIGHT_PAIR):    3,
                (SC_FORWARD,    3,  LEFT_PAIR):     2,
                (SC_FORWARD,    3,  RIGHT_PAIR):    1}
    # return spot column
    return geopandas.pd.Series(zip(df['sc_orient'], df['track'], df['pair'])).map(map_spot).values

#
# Get Ancillary Field Name
#
def __getancillaryfield(parm, field_rec):
    if field_rec['anc_type'] == 0:
        return parm['atl03_ph_fields'][field_rec['field_index']]
    if field_rec['anc_type'] == 1:
        return parm['atl03_geo_fields'][field_rec['field_index']]
    if field_rec['anc_type'] == 2:
        return parm['atl08_fields'][field_rec['field_index']]
    if field_rec['anc_type'] == 3:
        return parm['atl06_fields'][field_rec['field_index']]
    raise sliderule.FatalError(f'Invalid ancillary field type {field_rec["anc_type"]}')

#
# Flatten Batches
#
def __flattenbatches(rsps, rectype, batch_column, parm, keep_id, as_numpy_array, height_key):

    # Check for Responses
    if rsps == None:
        return sliderule.emptyframe(crs=ICESAT2_CRS)

    # Check for Output Options
    if "output" in parm:
        gdf = sliderule.procoutputfile(parm, rsps)
        return gdf

    # Flatten Records
    columns = {}
    records = []
    num_records = 0
    field_dictionary = {} # [<field_name>] = {"extent_id": [], <field_name>: []}
    file_dictionary = {} # [id] = "filename"
    if len(rsps) > 0:
        # Sort Records
        for rsp in rsps:
            if rectype in rsp['__rectype']:
                records += rsp,
                num_records += len(rsp[batch_column])
            elif 'ancfrec' == rsp['__rectype']:
                for field_rec in rsp['fields']:
                    extent_id = numpy.uint64(rsp['extent_id'])
                    field_name = __getancillaryfield(parm, field_rec)
                    if field_name not in field_dictionary:
                        field_dictionary[field_name] = {'extent_id': [], field_name: []}
                    field_dictionary[field_name]['extent_id'] += extent_id,
                    field_dictionary[field_name][field_name] += sliderule.getvalues(field_rec['value'], field_rec['datatype'], len(field_rec['value']), num_elements=1)[0],
            elif 'rsrec' == rsp['__rectype'] or 'zsrec' == rsp['__rectype'] or 'sdrec' == rsp['__rectype']:
                if rsp["num_samples"] <= 0:
                    continue
                # Get field names and set
                sample = rsp["samples"][0]
                field_names = list(sample.keys())
                field_names.remove("__rectype")
                field_set = rsp['key']
                if rsp["num_samples"] > 1:
                    as_numpy_array = True
                # On first time, build empty dictionary for field set associated with raster
                if field_set not in field_dictionary:
                    field_dictionary[field_set] = {'extent_id': []}
                    for field in field_names:
                        field_dictionary[field_set][field_set + "." + field] = []
                # Populate dictionary for field set
                field_dictionary[field_set]['extent_id'] += numpy.uint64(rsp['index']),
                for field in field_names:
                    if as_numpy_array:
                        data = []
                        for s in rsp["samples"]:
                            data += s[field],
                        field_dictionary[field_set][field_set + "." + field] += numpy.array(data),
                    else:
                        field_dictionary[field_set][field_set + "." + field] += sample[field],
            elif 'waverec' == rsp['__rectype']:
                field_set = rsp['__rectype']
                field_names = list(rsp.keys())
                field_names.remove("__rectype")
                if field_set not in field_dictionary:
                    field_dictionary[field_set] = {'extent_id': []}
                    for field in field_names:
                        field_dictionary[field_set][field] = []
                for field in field_names:
                    if type(rsp[field]) == tuple:
                        field_dictionary[field_set][field] += numpy.array(rsp[field]),
                    elif field == 'extent_id':
                        field_dictionary[field_set][field] += numpy.uint64(rsp[field]),
                    else:
                        field_dictionary[field_set][field] += rsp[field],
            elif 'fileidrec' == rsp['__rectype']:
                file_dictionary[rsp["file_id"]] = rsp["file_name"]

        # Build Columns
        if num_records > 0:
            # Initialize Columns
            sample_record = records[0][batch_column][0]
            for field in sample_record.keys():
                fielddef = sliderule.getdefinition(sample_record['__rectype'], field)
                if len(fielddef) > 0:
                    if type(sample_record[field]) == tuple:
                        columns[field] = numpy.empty(num_records, dtype=object)
                    else:
                        columns[field] = numpy.empty(num_records, fielddef["nptype"])
            # Populate Columns
            cnt = 0
            for record in records:
                for batch in record[batch_column]:
                    for field in columns:
                        columns[field][cnt] = batch[field]
                    cnt += 1
    else:
        logger.debug("No response returned")

    # Build Initial GeoDataFrame
    gdf = sliderule.todataframe(columns, height_key=height_key, crs=ICESAT2_CRS)

    # Merge Ancillary Fields
    for field_set in field_dictionary:
        df = geopandas.pd.DataFrame(field_dictionary[field_set])
        gdf = geopandas.pd.merge(gdf, df, how='left', on='extent_id').set_axis(gdf.index)

    # Delete Extent ID Column
    if len(gdf) > 0 and not keep_id:
        del gdf["extent_id"]

    # Attach Metadata
    if len(file_dictionary) > 0:
        gdf.attrs['file_directory'] = file_dictionary

    # Return GeoDataFrame
    return gdf

#
# Flatten Batches - ATL03
#
def __flattenbatches03(rsps, parm, keep_id, height_key):

    # Check for Responses
    if rsps == None:
        return sliderule.emptyframe(crs=ICESAT2_CRS)

    # Check for Output Options
    if "output" in parm:
        return sliderule.procoutputfile(parm, rsps)

    else: # Native Output
        # Flatten Responses
        columns = {}
        sample_photon_record = None
        photon_records = []
        num_photons = 0
        photon_dictionary = {}
        photon_field_types = {} # ['field_name'] = nptype
        if len(rsps) > 0:
            # Sort Records
            for rsp in rsps:
                if 'atl03rec' in rsp['__rectype']:
                    photon_records += rsp,
                    num_photons += len(rsp['photons'])
                    if sample_photon_record == None and len(rsp['photons']) > 0:
                        sample_photon_record = rsp
                elif 'ancerec' == rsp['__rectype']:
                    # Get Field Name and Type
                    field_name = __getancillaryfield(parm, rsp)
                    if field_name not in photon_field_types:
                        photon_field_types[field_name] = BASIC_TYPES[CODED_TYPE[rsp['datatype']]]["nptype"]
                    # Initialize Extent Dictionary Entry
                    extent_id = rsp['extent_id']
                    if extent_id not in photon_dictionary:
                        photon_dictionary[extent_id] = {}
                    # Save of Values per Extent ID per Field Name
                    data = sliderule.getvalues(rsp['data'], rsp['datatype'], len(rsp['data']))
                    photon_dictionary[extent_id][field_name] = data
            # Build Elevation Columns
            if num_photons > 0:
                # Initialize Columns
                for field in sample_photon_record.keys():
                    fielddef = sliderule.getdefinition("atl03rec", field)
                    if len(fielddef) > 0 and field != "photons":
                        columns[field] = numpy.empty(num_photons, fielddef["nptype"])
                for field in sample_photon_record["photons"][0].keys():
                    fielddef = sliderule.getdefinition("atl03rec.photons", field)
                    if len(fielddef) > 0:
                        columns[field] = numpy.empty(num_photons, fielddef["nptype"])
                for field in photon_field_types.keys():
                    columns[field] = numpy.empty(num_photons, photon_field_types[field])
                # Populate Columns
                ph_cnt = 0
                for record in photon_records:
                    # Add Ancillary Extent Fields
                    ph_index = 0
                    extent_id = record['extent_id']
                    if extent_id in photon_dictionary:
                        for photon in record["photons"]:
                            for field_name, field_array in photon_dictionary[extent_id].items():
                                columns[field_name][ph_cnt + ph_index] = field_array[ph_index]
                            ph_index += 1
                    # For Each Photon in Extent
                    for photon in record["photons"]:
                        # Add per Extent Fields
                        for field in record.keys():
                            if field in columns:
                                columns[field][ph_cnt] = record[field]
                        # Add per Photon Fields
                        for field in photon.keys():
                            if field in columns:
                                columns[field][ph_cnt] = photon[field]
                        # Goto Next Photon
                        ph_cnt += 1

                # Delete Extent ID Column
                if "extent_id" in columns and not keep_id:
                    del columns["extent_id"]

                # Create DataFrame
                gdf = sliderule.todataframe(columns, height_key=height_key, crs=ICESAT2_CRS)

                # Calculate Spot Column
                gdf['spot'] = __calcspot(gdf)

                # Return Response
                return gdf
            else:
                logger.debug("No photons returned")
        else:
            logger.debug("No response returned")

    # Return Empty Response
    return sliderule.emptyframe(crs=ICESAT2_CRS)

#
# Flatten Batches - ATL03 Viewer
#
def __flattenbatches03v(rsps, parm, keep_id):

    # Check for Responses
    if rsps == None:
        return sliderule.emptyframe(crs=ICESAT2_CRS)

    # Check for Output Options
    if "output" in parm:
        return sliderule.procoutputfile(parm, rsps)

    else: # Native Output
        # Flatten Responses
        columns = {}
        sample_segment_record = None
        extent_records = []
        num_segments = 0
        if len(rsps) > 0:
            # Sort Records
            for rsp in rsps:
                if 'atl03vrec' in rsp['__rectype']:
                    extent_records += rsp,
                    num_segments += len(rsp['segments'])
                    if sample_segment_record == None and len(rsp['segments']) > 0:
                        sample_segment_record = rsp
            # Build Segments Columns
            if num_segments > 0:
                # Initialize Columns
                for field in sample_segment_record.keys():
                    fielddef = sliderule.getdefinition("atl03vrec", field)
                    if len(fielddef) > 0 and field != "segments":
                        columns[field] = numpy.empty(num_segments, fielddef["nptype"])
                for field in sample_segment_record["segments"][0].keys():
                    fielddef = sliderule.getdefinition("atl03vrec.segments", field)
                    if len(fielddef) > 0:
                        columns[field] = numpy.empty(num_segments, fielddef["nptype"])
                # Populate Columns
                seg_cnt = 0
                for record in extent_records:
                    # For Each Segment in Extent
                    for segment in record["segments"]:
                        # Add per Extent Fields
                        for field in record.keys():
                            if field in columns:
                                columns[field][seg_cnt] = record[field]
                        # Add per Segments Fields
                        for field in segment.keys():
                            if field in columns:
                                columns[field][seg_cnt] = segment[field]
                        # Goto Next Segment
                        seg_cnt += 1

                # Delete Extent ID Column
                if "extent_id" in columns and not keep_id:
                    del columns["extent_id"]

                # Create DataFrame
                gdf = sliderule.todataframe(columns, crs=ICESAT2_CRS)

                # Return Response
                return gdf
            else:
                logger.debug("No segment counts returned")
        else:
            logger.debug("No response returned")

    # Error Case
    return sliderule.emptyframe(crs=ICESAT2_CRS)

#
# Build Request
#
def __build_request(parm, resources, default_asset='icesat2'):

    # Default the Asset
    rqst_parm = parm.copy()
    if "asset" not in rqst_parm:
        rqst_parm["asset"] = default_asset

    # Build Request
    return {
        "resources": resources,
        "parms": rqst_parm
    }



import os
import sys
import json
import time
import argparse
import duckdb
import geopandas as gpd
from shapely.geometry import mapping, Point

# -------------------------------------------
# command line arguments
# -------------------------------------------
parser = argparse.ArgumentParser(description="""ATL24""")
parser.add_argument('--atl13_shapefile',    type=str,               default="/data/ATL13/bodyfile/ATL13_Inland_Water_Body_Mask_v3_20191220.shp")
parser.add_argument('--atl13_parquet',      type=str,               default="/data/ATL13/inland_water_body.parquet")
parser.add_argument('--atl13_duckdb',       type=str,               default="/data/ATL13/inland_water_body.db")
parser.add_argument('--atl13_mappings',     type=str,               default="/data/ATL13/atl13_mappings.json")
parser.add_argument('--convert_shapefile',  action='store_true',    default=False)
parser.add_argument('--display_duckdb',     action='store_true',    default=False)
parser.add_argument('--use_parquet',        action='store_true',    default=False)
parser.add_argument('--entry_by_refid',     type=int,               default=None) # 5952002394
parser.add_argument('--entry_by_name',      type=str,               default=None) # "Caspian Sea"
parser.add_argument('--entry_by_coord',     nargs='+', type=float,  default=[]) # -86.79835088109307 42.762733124439904
args,_ = parser.parse_known_args()

# -------------------------------------------
# load parquet
# -------------------------------------------
def load_parquet():
    # read mappings
    with open(args.atl13_mappings, "r") as file:
        print(f'Loading ATL13 mappings... ', end='')
        sys.stdout.flush()
        start_time = time.perf_counter()
        atl13_mappings = json.loads(file.read())
        print(f'completed in {time.perf_counter() - start_time:.2f} secs.')
    # read body mask
    print(f'Loading ATL13 body mask... ', end='')
    sys.stdout.flush()
    start_time = time.perf_counter()
    atl13_mask = gpd.read_parquet(args.atl13_parquet)
    print(f'completed in {time.perf_counter() - start_time:.2f} secs.')
    # return parquet and mappings
    return atl13_mask, atl13_mappings

# -------------------------------------------
# load database
# -------------------------------------------
def load_database():
    # read mappings
    with open(args.atl13_mappings, "r") as file:
        print(f'Loading ATL13 mappings... ', end='')
        sys.stdout.flush()
        start_time = time.perf_counter()
        atl13_mappings = json.loads(file.read())
        print(f'completed in {time.perf_counter() - start_time:.2f} secs.')
    # read body mask
    print(f'Loading ATL13 body mask... ', end='')
    sys.stdout.flush()
    start_time = time.perf_counter()
    atl13_mask = duckdb.connect(args.atl13_duckdb)
    atl13_mask.execute("LOAD spatial;")
    print(f'completed in {time.perf_counter() - start_time:.2f} secs.')
    # return database and mappings
    return atl13_mask, atl13_mappings

# -------------------------------------------
# display_entry
# -------------------------------------------
def display_entry(refid):
    print("\nEntry\n-----\n", atl13_mask.loc[refid].to_dict(), "\n")
    print("Geometry\n--------\n", mapping(atl13_mask.loc[refid].geometry), "\n")
    print("Granules\n--------")
    for granule_id in atl13_mappings["refids"][str(refid)]:
        print(atl13_mappings["granules"][str(granule_id)])

# -------------------------------------------
# convert shapefile to parquet & duckdb
# -------------------------------------------
if args.convert_shapefile:
    print(f'Reading shapefile {args.atl13_shapefile}')
    gdf = gpd.read_file(args.atl13_shapefile)
    gdf["ATL13refID"] = gdf["ATL13refID"].astype(int)
    gdf.set_index("ATL13refID", inplace=True)
    print(f'Writing parquet file {args.atl13_parquet}')
    gdf.to_parquet(args.atl13_parquet, index=True)
    try:
        os.remove(args.atl13_duckdb)
        print(f'Replacing old database {args.atl13_duckdb}')
    except:
        print(f'Creating new database {args.atl13_duckdb}')
    db = duckdb.connect(args.atl13_duckdb)
    db.execute(f"""
        INSTALL spatial;
        LOAD spatial;
        CREATE TABLE atl13_mask AS
        SELECT *
        FROM '{args.atl13_parquet}'
        ORDER BY ATL13refID;
    """)
    print(f'Complete.')

# -------------------------------------------
# display structure of duckdb database
# -------------------------------------------
if args.display_duckdb:
    db = duckdb.connect(args.atl13_duckdb)
    df = db.execute("DESCRIBE atl13_mask").fetchdf()
    print(df)

# -------------------------------------------
# get entry by reference id
# -------------------------------------------
if args.entry_by_refid != None:
    if args.use_parquet:
        atl13_mask, atl13_mappings = load_parquet()
        display_entry(args.entry_by_refid)
    else:
        atl13_mask, atl13_mappings = load_database()
        df = atl13_mask.execute(f"""
            SELECT *
            FROM atl13_mask
            WHERE ATL13refID == {args.entry_by_refid};
        """).df()
        print(df)

# -------------------------------------------
# get entry by lake name
# -------------------------------------------
if args.entry_by_name != None:
    if args.use_parquet:
        atl13_mask, atl13_mappings = load_parquet()
        atl13_mask = atl13_mask[atl13_mask["Lake_name"] == args.entry_by_name]
        display_entry(atl13_mask.index[0])
    else:
        atl13_mask, atl13_mappings = load_database()
        df = atl13_mask.execute(f"""
            SELECT *
            FROM atl13_mask
            WHERE Lake_name == '{args.entry_by_name}';
        """).df()
        print(df)

# -------------------------------------------
# get entry by coordinates
# -------------------------------------------
if len(args.entry_by_coord) == 2:
    if args.use_parquet:
        atl13_mask, atl13_mappings = load_parquet()
        point = Point(args.entry_by_coord) # longitude, latitude
        atl13_mask = atl13_mask[atl13_mask.geometry.contains(point)]
        display_entry(atl13_mask.index[0])
    else:
        atl13_mask, atl13_mappings = load_database()
        df = atl13_mask.execute(f"""
            SELECT *
            FROM atl13_mask
            WHERE ST_Contains(geometry, ST_Point({args.entry_by_coord[0]}, {args.entry_by_coord[1]}));
        """).df()
        print(df)



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import itertools
import copy
import json
import ssl
import urllib.request
import http
import requests
import numpy
import geopandas
from datetime import datetime
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry import Polygon
from sliderule import sliderule, logger

###############################################################################
# GLOBALS
###############################################################################

# profiling times for each major function
profiles = {}

# default maximum number of resources to process in one request
DEFAULT_MAX_REQUESTED_RESOURCES = 300
max_requested_resources = DEFAULT_MAX_REQUESTED_RESOURCES

# best effort match of datasets to providers and versions for earthdata
DATASETS = {
    "ATL03":                                               {"provider": "NSIDC_CPRD",  "version": "006",  "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "ATL06":                                               {"provider": "NSIDC_CPRD",  "version": "006",  "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "ATL08":                                               {"provider": "NSIDC_CPRD",  "version": "006",  "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "ATL09":                                               {"provider": "NSIDC_CPRD",  "version": "006",  "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "ATL13":                                               {"provider": "NSIDC_CPRD",  "version": "006",  "api": "cmr",   "formats": {".h5"},    "collections": [],                                 "url": None},
    "GEDI01_B":                                            {"provider": "LPCLOUD",     "version": "002",  "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "GEDI02_A":                                            {"provider": "LPCLOUD",     "version": "002",  "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "GEDI_L3_LandSurface_Metrics_V2_1952":                 {"provider": "ORNL_CLOUD",  "version": None,   "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "GEDI_L4A_AGB_Density_V2_1_2056":                      {"provider": "ORNL_CLOUD",  "version": None,   "api": "cmr",   "formats": [".h5"],    "collections": [],                                 "url": None},
    "GEDI_L4B_Gridded_Biomass_2017":                       {"provider": "ORNL_CLOUD",  "version": None,   "api": "cmr",   "formats": [".tiff"],  "collections": [],                                 "url": None},
    "HLS":                                                 {"provider": "LPCLOUD",     "version": None,   "api": "stac",  "formats": [".tiff"],  "collections": ["HLSS30.v2.0", "HLSL30.v2.0"],     "url": 'https://cmr.earthdata.nasa.gov/stac/LPCLOUD/search'},
    "Digital Elevation Model (DEM) 1 meter":               {"provider": "USGS",        "version": None,   "api": "tnm",   "formats": [".tiff"],  "collections": [],                                 "url": None},
    "SWOT_SIMULATED_L2_KARIN_SSH_ECCO_LLC4320_CALVAL_V1":  {"provider": "POCLOUD",     "version": None,   "api": "cmr",   "formats": [".nc"],    "collections": ["C2147947806-POCLOUD"],            "url": None},
    "SWOT_SIMULATED_L2_KARIN_SSH_GLORYS_CALVAL_V1":        {"provider": "POCLOUD",     "version": None,   "api": "cmr",   "formats": [".nc"],    "collections": ["C2152046451-POCLOUD"],            "url": None},
    "arcticdem-strips":                                    {"provider": "PGC",         "version": None,   "api": "stac",  "formats": [".tiff"],  "collections": ["arcticdem-strips-s2s041-2m"],     "url": 'https://stac.pgc.umn.edu/api/v1/search'},
    "rema-strips":                                         {"provider": "PGC",         "version": None,   "api": "stac",  "formats": [".tiff"],  "collections": ["rema-strips-s2s041-2m"],          "url": 'https://stac.pgc.umn.edu/api/v1/search'}
}

# best effort match of sliderule assets to earthdata datasets
ASSETS_TO_DATASETS = {
    "gedil4a": "GEDI_L4A_AGB_Density_V2_1_2056",
    "gedil4b": "GEDI_L4B_Gridded_Biomass_2017",
    "gedil3-elevation": "GEDI_L3_LandSurface_Metrics_V2_1952",
    "gedil3-canopy": "GEDI_L3_LandSurface_Metrics_V2_1952",
    "gedil3-elevation-stddev": "GEDI_L3_LandSurface_Metrics_V2_1952",
    "gedil3-canopy-stddev": "GEDI_L3_LandSurface_Metrics_V2_1952",
    "gedil3-counts": "GEDI_L3_LandSurface_Metrics_V2_1952",
    "gedil2a": "GEDI02_A",
    "gedil1b": "GEDI01_B",
    "swot-sim-ecco-llc4320": "SWOT_SIMULATED_L2_KARIN_SSH_ECCO_LLC4320_CALVAL_V1",
    "swot-sim-glorys": "SWOT_SIMULATED_L2_KARIN_SSH_GLORYS_CALVAL_V1",
    "usgs3dep-1meter-dem": "Digital Elevation Model (DEM) 1 meter",
    "landsat-hls": "HLS",
    "icesat2": "ATL03",
    "icesat2-atl06": "ATL06",
    "icesat2-atl08": "ATL08",
    "icesat2-atl09": "ATL09",
    "icesat2-atl13": "ATL13",
    "atlas-local": "ATL03",
    "nsidc-s3": "ATL03",
    "arcticdem-strips": "arcticdem-strips",
    "rema-strips": "rema-strips"
}

# upper limit on resources returned from CMR query per request
CMR_PAGE_SIZE = 2000

# upper limit on resources returned from STAC query per request
STAC_PAGE_SIZE = 100

# upper limit on resources returned from TNM query per request
TNM_PAGE_SIZE = 100

###############################################################################
# NSIDC UTILITIES
###############################################################################
# The functions below have been adapted from the NSIDC download script and
# carry the following notice:
#
# Copyright (c) 2020 Regents of the University of Colorado
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

def __cmr_filter_urls(search_results, data_formats):
    """Select only the desired data files from CMR response."""
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return []
    entries = [e['links']
               for e in search_results['feed']['entry']
               if 'links' in e]
    # Flatten "entries" to a simple list of links
    links = list(itertools.chain(*entries))
    # Build unique filenames
    urls = []
    unique_filenames = set()
    for link in links:
        if 'href' not in link:
            # Exclude links with nothing to download
            continue
        if 'inherited' in link and link['inherited'] is True:
            # Why are we excluding these links?
            continue
        if 'rel' in link and 'data#' not in link['rel']:
            # Exclude links which are not classified by CMR as "data" or "metadata"
            continue
        if 'title' in link and 'opendap' in link['title'].lower():
            # Exclude OPeNDAP links--they are responsible for many duplicates
            # This is a hack; when the metadata is updated to properly identify
            # non-datapool links, we should be able to do this in a non-hack way
            continue
        filename = link['href'].split('/')[-1]
        if filename in unique_filenames:
            # Exclude links with duplicate filenames (they would overwrite)
            continue
        unique_filenames.add(filename)
        if any([link['href'].endswith(extension) for extension in data_formats]):
            resource = link['href'].split("/")[-1]
            urls.append(resource)
    # return filtered urls
    return urls

def __cmr_granule_metadata(search_results, crs=sliderule.DEFAULT_CRS):
    """Get the metadata for CMR returned granules"""
    # GeoDataFrame with granule metadata
    granule_metadata = sliderule.emptyframe(crs=crs)
    # return empty dataframe if no CMR entries
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return granule_metadata
    # for each CMR entry
    for e in search_results['feed']['entry']:
        # columns for dataframe
        columns = {}
        # granule title and identifiers
        columns['title'] = e['title']
        columns['collection_concept_id'] = e['collection_concept_id']
        # time start and time end of granule
        columns['time_start'] = numpy.datetime64(e['time_start'])
        columns['time_end'] = numpy.datetime64(e['time_end'])
        columns['time_updated'] = numpy.datetime64(e['updated'])
        # get the granule size and convert to bits
        columns['granule_size'] = float(e['granule_size'])*(2.0**20)
        # Create Pandas DataFrame object
        # use granule id as index
        df = geopandas.pd.DataFrame(columns, index=[e['id']])
        # Generate Geometry Column
        if 'polygons' in e:
            polygons = []
            # for each polygon
            for poly in e['polygons'][0]:
                coords = [float(i) for i in poly.split()]
                polygons.append(Polygon(zip(coords[1::2], coords[::2])))
            # generate multipolygon from list of polygons
            geometry = MultiPolygon(polygons)
        else:
            geometry, = geopandas.points_from_xy([None], [None])
        # Build GeoDataFrame (default geometry is crs=EPSG_MERCATOR)
        gdf = geopandas.GeoDataFrame(df, geometry=[geometry], crs=crs)
        # append to combined GeoDataFrame and catch warnings
        granule_metadata = geopandas.pd.concat([granule_metadata, gdf])
    # return granule metadata
    # - time start and time end
    # - time granule was updated
    # - granule size in bits
    # - polygons as geodataframe geometry
    return granule_metadata

def __cmr_collection_query(provider, short_name):
    """Perform a CMR query for collection metadata."""
    # build params
    cmr_query_type = 'collections'
    cmr_format = 'json'
    CMR_URL = 'https://cmr.earthdata.nasa.gov'
    cmr_query_url = '/'.join([CMR_URL, 'search', f'{cmr_query_type}.{cmr_format}'])
    # build list of CMR query parameters
    params = []
    params.append(f'?provider={provider}')
    params.append(f'&short_name={short_name}')
    # full CMR query url
    cmr_query_url += "".join(params)
    logger.debug(f'cmr request={cmr_query_url}\n')
    # ssl context
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    # query CMR for collection metadata
    req = urllib.request.Request(cmr_query_url)
    response = urllib.request.urlopen(req, context=ctx)
    # read the CMR search as JSON
    search_results = json.loads(response.read().decode('utf8'))
    # return only valid collection entries
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return []
    else:
        return search_results['feed']['entry']

def __cmr_query(provider, short_name, version, time_start, time_end, **kwargs):
    """Perform a search-after CMR query for files matching input criteria."""
    kwargs.setdefault('polygon',None)
    kwargs.setdefault('name_filter',None)
    kwargs.setdefault('return_metadata',False)
    kwargs.setdefault('crs',sliderule.DEFAULT_CRS)
    # build params
    params = '&short_name={0}'.format(short_name)
    if version != None:
        params += '&version={0}'.format(version)
    if time_start is not None and time_end is not None:
        params += '&temporal[]={0},{1}'.format(time_start, time_end)
    if kwargs['polygon']:
        params += '&polygon={0}'.format(kwargs['polygon'])
    if kwargs['name_filter']:
        params += '&options[producer_granule_id][pattern]=true'
        params += '&producer_granule_id[]=' + kwargs['name_filter']

    CMR_URL = 'https://cmr.earthdata.nasa.gov'
    cmr_query_url = ('{0}/search/granules.json?provider={1}'
                     '&sort_key[]=start_date&sort_key[]=producer_granule_id'
                     '&page_size={2}'.format(CMR_URL, provider, CMR_PAGE_SIZE))
    cmr_query_url += params
    logger.debug(f'Initial CMR request: {cmr_query_url}')

    cmr_search_after = None
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    urls = []
    # GeoDataFrame with granule metadata
    metadata = sliderule.emptyframe(crs=kwargs['crs'])
    while True:
        req = urllib.request.Request(cmr_query_url)
        if cmr_search_after:
            req.add_header('CMR-Search-After', cmr_search_after)
            logger.debug(f'Requesting next page with CMR-Search-After: {cmr_search_after}')

        response = urllib.request.urlopen(req, context=ctx)

        headers = {k.lower(): v for k, v in dict(response.info()).items()}
        cmr_search_after = headers.get('cmr-search-after')

        search_page = response.read()
        search_page = json.loads(search_page.decode('utf-8'))

        url_scroll_results = __cmr_filter_urls(search_page, DATASETS[short_name]["formats"])
        if not url_scroll_results:
            break
        urls += url_scroll_results
        # query for granule metadata and polygons
        if kwargs['return_metadata']:
            metadata_results = __cmr_granule_metadata(search_page, crs=kwargs['crs'])
        else:
            metadata_results = geopandas.pd.DataFrame([None for _ in url_scroll_results])

        # append granule metadata
        metadata = geopandas.pd.concat([metadata, metadata_results])

        # Two ways to determine that there is no more data available:
        # 1. The number of granules in the current response is less than the requested 'page_size':
        # 2. The absence of the 'CMR-Search-After' header
        result_count = len(search_page['feed']['entry'])
        if result_count < CMR_PAGE_SIZE:
            logger.debug(f'Received {result_count} results, fewer than page size. Ending pagination after processing.')
            break
        if not cmr_search_after:
            logger.debug('No CMR-Search-After header found, no more pages.')
            break

    return urls, metadata

###############################################################################
# CMR UTILITIES
###############################################################################

#
# Perform a CMR Search Request
#
def __cmr_search(short_name, version, polygons, time_start, time_end, return_metadata, name_filter):

    global max_requested_resources

    tstart = time.perf_counter()

    # check parameters
    if short_name == None:
        raise sliderule.FatalError("Must supply short name to CMR query")
    elif short_name not in DATASETS:
        raise sliderule.FatalError("Must supply a supported dataset: " + short_name)

    # attempt to fill in provider
    provider = DATASETS[short_name]["provider"]
    if provider == None:
        raise sliderule.FatalError("Unable to determine provider for CMR query")

    # attempt to fill in version
    if version == None:
        version = DATASETS[short_name]["version"]

    # initialize return value
    resources = {} # [<url>] = <polygon>

    # iterate through each polygon (or none if none supplied)
    for polygon in polygons:
        urls = []
        metadata = sliderule.emptyframe()

        # issue CMR request
        for tolerance in [0.0001, 0.001, 0.01, 0.1, 1.0, None]:

            # convert polygon list into string
            polystr = None
            if polygon:
                flatpoly = []
                for p in polygon:
                    flatpoly.append(p["lon"])
                    flatpoly.append(p["lat"])
                polystr = str(flatpoly)[1:-1]
                polystr = polystr.replace(" ", "") # remove all spaces as this will be embedded in a url

            # call into NSIDC routines to make CMR request
            try:
                urls,metadata = __cmr_query(provider, short_name, version, time_start, time_end, polygon=polystr, return_metadata=return_metadata, name_filter=name_filter)
                break # exit loop because cmr search was successful
            except urllib.error.HTTPError as e:
                logger.error('HTTP Request Error: {}'.format(e))
            except http.client.HTTPException as e:
                logger.error('HTTP Client Error: {}'.format(e))
            except RuntimeError as e:
                logger.error('Runtime Error: {}'.format(e))

            # simplify polygon
            if polygon and tolerance:
                raw_multi_polygon = [[(tuple([(c['lon'], c['lat']) for c in polygon]), [])]]
                shape = MultiPolygon(*raw_multi_polygon)
                buffered_shape = shape.buffer(tolerance)
                simplified_shape = buffered_shape.simplify(tolerance)
                simplified_coords = list(simplified_shape.exterior.coords)
                logger.warning('Using simplified polygon (for CMR request only!), {} points using tolerance of {}'.format(len(simplified_coords), tolerance))
                region = []
                for coord in simplified_coords:
                    point = {"lon": coord[0], "lat": coord[1]}
                    region.insert(0,point)
                polygon = region
            else:
                break # exit here because nothing can be done

        # populate resources
        for i,url, in enumerate(urls):
            resources[url] = metadata.iloc[i]

    # build return lists
    url_list = list(resources.keys())
    meta_list = list(resources.values())

    # Check Resources are Under Limit
    if(len(url_list) > max_requested_resources):
        raise sliderule.FatalError('Exceeded maximum requested resources: {} (current max is {})\nConsider using earthdata.set_max_resources to set a higher limit.'.format(len(url_list), max_requested_resources))
    else:
        logger.info("Identified %d resources to process", len(url_list))

    profiles[__cmr_search.__name__] = time.perf_counter() - tstart

    if return_metadata:
        return (url_list,meta_list)
    else:
        return url_list

#
# Get the maximum available version for a dataset from CMR
#
def __cmr_max_version(provider, short_name):
    """Get the maximum version of a dataset from CMR."""
    entries = __cmr_collection_query(provider, short_name)
    version_ids = [e['version_id'] for e in entries]
    # return only when there was valid collection entries
    if not version_ids:
        return None
    else:
        return max(version_ids)

#
# Build a GeoJSON Response from STAC Query Response
#
def __build_geojson(rsps):
    geojson = rsps.json()
    next_page = {"link": None, "body": None}
    if "links" in geojson:
        for link in geojson["links"]:
            if link["rel"] == "next":
                next_page["link"] = link["href"]
                if "body" in link:
                    next_page["body"] = json.dumps(link["body"]) # POST based pagination
                break
        del geojson["links"]
    if 'numberMatched' in geojson:
        del geojson['numberMatched']
    if 'numberReturned' in geojson:
        del geojson['numberReturned']
    for i in reversed(range(len(geojson["features"]))):
        del geojson["features"][i]["links"]
        del geojson["features"][i]["stac_version"]
        del geojson["features"][i]["stac_extensions"]
        del geojson["features"][i]["collection"]
        del geojson["features"][i]["bbox"]
        if "browse" in geojson["features"][i]["assets"]:
            del geojson["features"][i]["assets"]["browse"]
        if "metadata" in geojson["features"][i]["assets"]:
            del geojson["features"][i]["assets"]["metadata"]
        assetsDict = geojson["features"][i]["assets"]
        for val in assetsDict:
            if "href" in assetsDict[val]:
                geojson["features"][i]["properties"][val] = assetsDict[val]["href"]
        del geojson["features"][i]["assets"]

    return geojson, next_page

#
# Perform a STAC Query
#
#   See https://cmr.earthdata.nasa.gov/stac/docs/index.html for details on API
#
def __stac_search(short_name, collections, polygons, time_start, time_end):

    # check short name
    if short_name not in DATASETS:
        raise sliderule.FatalError(f'{short_name} is not a supported dataset for STAC queries')

    # create requests context
    context = requests.Session()
    context.trust_env = False # prevent requests from attaching credentials from environment

    # build stac request
    url = DATASETS[short_name]["url"]
    headers = {'Content-Type': 'application/json'}
    rqst = {
        "limit": STAC_PAGE_SIZE,
        "datetime": '{}/{}'.format(time_start, time_end),
        "collections": collections or DATASETS[short_name]["collections"],
    }

    # add polygon if provided
    if polygons:
        if len(polygons) == 1 and len(polygons[0]) == 1:
            # Single point
            coord = polygons[0][0]
            rqst["intersects"] = {
                "type": "Point",
                "coordinates": [coord["lon"], coord["lat"]]
            }
        else:
            # Regular polygon handling
            rqst["intersects"] = {
                "type": "Polygon",
                "coordinates": [[[coord["lon"], coord["lat"]] for coord in polygon] for polygon in polygons]
            }

    # make initial stac request
    data = context.post(url, data=json.dumps(rqst), headers=headers)
    data.raise_for_status()
    geojson, next_page = __build_geojson(data)

    # Continue fetching pages if 'next' link is available
    while next_page["link"]:
        data = context.post(next_page["link"], data=next_page["body"], headers=headers)
        data.raise_for_status()
        _geojson, next_page = __build_geojson(data)
        geojson["features"] += _geojson["features"]

    # set number of returned features
    geojson["context"]["returned"] = len(geojson["features"])

    # return geojson dictionary
    return geojson

#
# Format Polygons for Request
#
def __format_polygons(polygon):

    polygons = [None]

    # create list of polygons
    if polygon and len(polygon) > 0:
        if type(polygon[0]) == dict:
            polygons = [copy.deepcopy(polygon)]
        elif type(polygon[0] == list):
            polygons = copy.deepcopy(polygon)

    # return list of polygons
    return polygons

###############################################################################
# APIs
###############################################################################

#
#  Set Maximum Resources
#
def set_max_resources (max_resources):
    '''
    Sets the maximum allowed number of resources to be processed in one request.  This is mainly provided as a sanity check for the user.

    Parameters
    ----------
        max_resources : int
                        the maximum number of resources that are allowed to be processed in a single request

    Examples
    --------
        >>> from sliderule import icesat2
        >>> earthdata.set_max_resources(1000)
    '''
    global max_requested_resources
    max_requested_resources = max_resources

#
#  Common Metadata Repository
#
def cmr(short_name=None, version=None, polygon=None, time_start='2018-01-01T00:00:00Z', time_end=datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"), return_metadata=False, name_filter=None):
    '''
    Query the `NASA Common Metadata Repository (CMR) <https://cmr.earthdata.nasa.gov>`_ for a list of data within temporal and spatial parameters

    Parameters
    ----------
        short_name:         str
                            dataset short name as defined in the `NASA CMR Directory <https://cmr.earthdata.nasa.gov/search/site/collections/directory/eosdis>`_
        version:            str
                            dataset version string, leave as None to get latest support version
        polygon:            list
                            either a single list of longitude,latitude in counter-clockwise order with first and last point matching, defining region of interest (see `polygons </web/rtd/user_guide/SlideRule.html#polygons>`_), or a list of such lists when the region includes more than one polygon
        time_start:         str
                            starting time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        time_end:           str
                            ending time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        return_metadata:    bool
                            flag indicating whether metadata associated with the query is returned back to the user
        name_filter:        str
                            filter to apply to resources returned by query

    Returns
    -------
    list
        files (granules) for the dataset fitting the spatial and temporal parameters

    Examples
    --------
        >>> from sliderule import earthdata
        >>> region = [ {"lon": -108.3435200747503, "lat": 38.89102961045247},
        ...            {"lon": -107.7677425431139, "lat": 38.90611184543033},
        ...            {"lon": -107.7818591266989, "lat": 39.26613714985466},
        ...            {"lon": -108.3605610678553, "lat": 39.25086131372244},
        ...            {"lon": -108.3435200747503, "lat": 38.89102961045247} ]
        >>> granules = earthdata.cmr(short_name='ATL06', polygon=region)
        >>> granules
        ['ATL03_20181017222812_02950102_003_01.h5', 'ATL03_20181110092841_06530106_003_01.h5', ... 'ATL03_20201111102237_07370902_003_01.h5']
    '''
    # create list of polygons
    polygons = __format_polygons(polygon)

    # perform query
    return __cmr_search(short_name, version, polygons, time_start, time_end, return_metadata, name_filter)

#
#  SpatioTemporal Asset Catalog
#
def stac(short_name=None, collections=None, polygon=None, time_start='2018-01-01T00:00:00Z', time_end=datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"), as_str=True):
    '''
    Perform a STAC query of the `NASA Common Metadata Repository (CMR) <https://cmr.earthdata.nasa.gov>`_ catalog for a list of data within temporal and spatial parameters

    Parameters
    ----------
        short_name:     str
                        dataset short name as defined in the `NASA CMR Directory <https://cmr.earthdata.nasa.gov/search/site/collections/directory/eosdis>`_
        collections:    list
                        list of dataset collections as specified by CMR, leave as None to use defaults
        polygon:        list
                        either a single list of longitude,latitude in counter-clockwise order with first and last point matching, defining region of interest (see `polygons </web/rtd/user_guide/SlideRule.html#polygons>`_), or a list of such lists when the region includes more than one polygon
        time_start:     str
                        starting time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        time_end:       str
                        ending time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        as_str:         bool
                        whether to return geojson as a dictionary or string

    Returns
    -------
    str
        geojson of the feature set returned by the query
        {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "id": "<id>",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [..]
                    },
                    "properties": {
                        "datetime": "YYYY-MM-DDTHH:MM:SS.SSSZ",
                        "start_datetime": "YYYY-MM-DDTHH:MM:SS.SSSZ",
                        "end_datetime": "YYYY-MM-DDTHH:MM:SS.SSSZ",
                        "<tag>": "<url>",
                        ..
                    }
                },
                ..
            ],
            "stac_version": "<version>"
            "context": {
                "returned": <returned>,
                "limit": <limit>,
                "matched": <matched>
            }
        }

    Examples
    --------
        >>> from sliderule import earthdata
        >>> region = [ {"lon": -108.3435200747503, "lat": 38.89102961045247},
        ...            {"lon": -107.7677425431139, "lat": 38.90611184543033},
        ...            {"lon": -107.7818591266989, "lat": 39.26613714985466},
        ...            {"lon": -108.3605610678553, "lat": 39.25086131372244},
        ...            {"lon": -108.3435200747503, "lat": 38.89102961045247} ]
        >>> geojson = earthdata.stac(short_name='HLS', polygon=region)
    '''
    # create list of polygons
    polygons = None
    if polygon:
        polygons = __format_polygons(polygon)

    # perform query
    geojson = __stac_search(short_name, collections, polygons, time_start, time_end)

    # return
    if as_str:
        return json.dumps(geojson)
    else:
        return geojson

#
#  The National Map
#
def tnm(short_name, polygon=None, time_start=None, time_end=datetime.utcnow().strftime("%Y-%m-%d"), as_str=True):
    '''
    Query `USGS National Map API <https://tnmaccess.nationalmap.gov/api/v1/products>`_ for a list of data within temporal and spatial parameters.  See https://apps.nationalmap.gov/help/documents/TNMAccessAPIDocumentation/TNMAccessAPIDocumentation.pdf for more details on the API.

    Parameters
    ----------
        short_name:         str
                            dataset name
        polygon:            list
                            a single list of longitude,latitude in counter-clockwise order with first and last point matching, defining region of interest (see `polygons </web/rtd/user_guide/SlideRule.html#polygons>`_)
        time_start:         str
                            starting time for query in format ``<year>-<month>-<day>``
        time_end:           str
                            ending time for query in format ``<year>-<month>-<day>``

    Returns
    -------
    dict
        geojson of resources intersecting area of interest

    Examples
    --------
        >>> from sliderule import earthdata
        >>> region = [ {"lon": -108.3435200747503, "lat": 38.89102961045247},
        ...            {"lon": -107.7677425431139, "lat": 38.90611184543033},
        ...            {"lon": -107.7818591266989, "lat": 39.26613714985466},
        ...            {"lon": -108.3605610678553, "lat": 39.25086131372244},
        ...            {"lon": -108.3435200747503, "lat": 38.89102961045247} ]
        >>> geojson = earthdata.tnm(short_name='Digital Elevation Model (DEM) 1 meter', polygon=region)
        >>> geojson
        {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'id': '5eaa4a0582cefae35a21ee8c', 'geometry': {'type': 'Polygon'...
    '''
    # Flatten polygon (the list must be formatted as 'x y, x y, x y, x y, x y', the documentation is incorrect)
    coord_list = []
    for coord in polygon:
        coord_list.append('{} {}'.format(coord["lon"], coord["lat"]))
    poly_str = ', '.join(coord_list)

    # Create requests context
    context = requests.Session()
    context.trust_env = False # prevent requests from attaching credentials from environment

    # Make paged requests to collect all resources that intersect area of interest
    items = []
    while True:

        # Build request
        url='https://tnmaccess.nationalmap.gov/api/v1/products'
        rqst = {
            "datasets":     short_name,
            "polygon":      poly_str,
            "prodFormats":  'GeoTIFF',
            "outputFormat": 'JSON',
            "max":          TNM_PAGE_SIZE,
            "offset":       len(items)
        }

        # Optional request parameters
        if time_start != None:
            rqst["start"] = time_start
            rqst["stop"] = time_end

        # Make request
        rsps = context.get(url, params=rqst)
        rsps.raise_for_status()
        try:
            data = json.loads(rsps.content)
        except:
            print(rsps.content)
            raise
        items += data['items']
        if len(items) == data['total']:
            break

    # Create GeoJSON
    geojson = {
        "type": "FeatureCollection",
        "features": []
    }
    for i in range(len(items)):
        geojson['features'].append({"type": "Feature"})
        id = items[i]['sourceId']
        geojson['features'][i].update({"id": id})

        # Create polygon from bounding box
        minX = items[i]['boundingBox']['minX']
        maxX = items[i]['boundingBox']['maxX']
        minY = items[i]['boundingBox']['minY']
        maxY = items[i]['boundingBox']['maxY']

        # Counter-clockwise for interior of bounding box
        coordList = [[[minX, minY], [maxX, minY], [maxX, maxY], [minX, maxY], [minX, minY]]]

        geometryDic = {"type": "Polygon", "coordinates": coordList}
        geojson['features'][i].update({"geometry": geometryDic})

        date = items[i]['lastUpdated']
        pydate = datetime.fromisoformat(date)
        date = pydate.astimezone(None).isoformat()
        url  = items[i]['urls']['TIFF']
        propertiesDict = {"datetime": date, "url": url }
        geojson['features'][i].update({"properties": propertiesDict})

    # Return GeoJSON
    if as_str:
        return json.dumps(geojson)
    else:
        return geojson


#
#  Search
#
def search(parm, resources=None):
    '''
    Return granules and populate catalogs for all required resources using the supplied parameters.  This is the highest-level API call and attempts to automatically determine which services need to be queried when making a request.

    Parameters
    ----------
        parm:               dict
                            request parameters
        resources:          list
                            list of granules to process

    Returns
    -------
    list
        list of granules to process

    Notes
    -----
    The ``parm`` argument will be modified in place.
    The ``asset`` parameter must be supplied

    Examples
    --------
        >>> from sliderule import earthdata
        >>> region = [ {"lon": -108.3435200747503, "lat": 38.89102961045247},
        ...            {"lon": -107.7677425431139, "lat": 38.90611184543033},
        ...            {"lon": -107.7818591266989, "lat": 39.26613714985466},
        ...            {"lon": -108.3605610678553, "lat": 39.25086131372244},
        ...            {"lon": -108.3435200747503, "lat": 38.89102961045247} ]
        >>> parms = {"asset": "icesat2", "poly": region, "cycle": 20, "rgt": 232}
        >>> resources = earthdata.search(parms)
    '''
    # Initialize Resources
    if resources == None:
        resources = []

    # Initialize CMR Keyword Arguments
    cmr_kwargs = {}

    # Pull Out Polygon
    if "ignore_poly_for_cmr" not in parm or not parm["ignore_poly_for_cmr"]:
        if "clusters" in parm and parm["clusters"] and len(parm["clusters"]) > 0:
            cmr_kwargs['polygon'] = parm["clusters"]
        elif "poly" in parm and parm["poly"] and len(parm["poly"]) > 0:
            cmr_kwargs['polygon'] = parm["poly"]

    # Pull Out Time Period
    if "t0" in parm:
        cmr_kwargs['time_start'] = parm["t0"]
    if "t1" in parm:
        cmr_kwargs['time_end'] = parm["t1"]

    # Build Name Filter
    name_filter_enabled = False
    if "asset" in parm:
        if "icesat2" in parm["asset"]:
            rgt_filter = '????'
            if "rgt" in parm and parm["rgt"] != None:
                rgt_filter = f'{parm["rgt"]}'.zfill(4)
                name_filter_enabled = True
            cycle_filter = '??'
            if "cycle" in parm and parm["cycle"] != None:
                cycle_filter = f'{parm["cycle"]}'.zfill(2)
                name_filter_enabled = True
            region_filter = '??'
            if "region" in parm and parm["region"] != None:
                region_filter = f'{parm["region"]}'.zfill(2)
                name_filter_enabled = True
            if name_filter_enabled:
                cmr_kwargs['name_filter'] = '*_' + rgt_filter + cycle_filter + region_filter + '_*'

    # Get Resources (CMR)
    if len(resources) == 0 and "asset" in parm:
        if (not name_filter_enabled) and ("poly" not in parm) and ("t0" not in parm) and ("t1" not in parm):
            logger.error("Must supply some bounding parameters with request (poly, t0, t1)")
            return resources
        elif parm["asset"] not in ASSETS_TO_DATASETS:
            logger.error(f'Unrecognized asset: {parm["asset"]}')
            return resources
        else:
            short_name = ASSETS_TO_DATASETS[parm["asset"]]
            resources = cmr(short_name=short_name, **cmr_kwargs)

    # Get Raster Catalogs (STAC, TNM)
    if "samples" in parm:
        for raster in parm["samples"].keys():
            sample_parm = parm["samples"][raster]
            if "catalog" not in sample_parm:
                try:
                    samples_kwargs = {"polygon": cmr_kwargs['polygon']}
                    dataset_name = ASSETS_TO_DATASETS[sample_parm["asset"]]
                    dataset = DATASETS[dataset_name]
                    # Pull Out Time Period
                    if "t0" in sample_parm:
                        samples_kwargs['time_start'] = sample_parm["t0"]
                    if "t1" in sample_parm:
                        samples_kwargs['time_end'] = sample_parm["t1"]
                    # Make GeoJSON Query
                    if dataset["api"] == "stac":
                        parm["samples"][raster]["catalog"] = stac(short_name=dataset_name, as_str=True, **samples_kwargs)
                    elif dataset["api"] == "tnm":
                        parm["samples"][raster]["catalog"] = tnm(short_name=dataset_name, as_str=True, **samples_kwargs)
                except Exception:
                    pass # best effort; many raster datasets don't need a catalog

    # Return Resources
    return resources



from pystac_client import Client
import sliderule
import requests
import json

# parameters
debug = False
client = False
manual = True
page = None
url = 'https://cmr.earthdata.nasa.gov/stac/LPCLOUD'
headers = {'Content-Type': 'application/json'}
limit = 200
collections = ["HLSS30.v2.0", "HLSL30.v2.0"]
time_start = "2021-01-01T00:00:00Z"
time_end = "2022-01-01T23:59:59Z"
coordinates = [[[coord["lon"], coord["lat"]] for coord in polygon] for polygon in [sliderule.toregion("examples/grandmesa.geojson")["poly"]]]

# debug output
if debug:
    import logging
    import http.client as http_client
    http_client.HTTPConnection.debuglevel = 1
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


# client stac request
if client:
    catalog = Client.open(url)
    timeRange = '{}/{}'.format(time_start, time_end)
    geometry = {"type": "Polygon", "coordinates": coordinates}
    results = catalog.search(collections=collections, datetime=timeRange, intersects=geometry)
    items = results.get_all_items_as_dict()
    print(f"Returned {results.matched()} features")


# manual stac request
if manual:
    rqst = {
        "limit": 100,
        "datetime": '{}/{}'.format(time_start, time_end),
        "collections": collections,
        "intersects": {
            "type": "Polygon",
            "coordinates": coordinates
        }
    }
    if page:
        rqst["page"] = page
    context = requests.Session()
    context.trust_env = False
    data = context.post(url+"/search", data=json.dumps(rqst), headers=headers)
    data.raise_for_status()
    geojson = data.json()
    print("Returned {} features".format(geojson["context"]["returned"]))




# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Imports
import sliderule
from sliderule import earthdata, icesat2, gedi
import argparse
import math
import numpy

# Command Line Arguments
parser = argparse.ArgumentParser(description="""Subset granules""")
parser.add_argument('--benchmarks',     '-b',   nargs='+', type=str,    default=[])
parser.add_argument('--domain',         '-d',   type=str,               default="slideruleearth.io")
parser.add_argument('--organization',   '-o',   type=str,               default="sliderule")
parser.add_argument('--desired_nodes',  '-n',   type=int,               default=None)
parser.add_argument('--time_to_live',   '-l',   type=int,               default=120)
parser.add_argument('--verbose',        '-v',   action='store_true',    default=False)
parser.add_argument('--loglvl',         '-j',   type=str,               default="CRITICAL")
args,_ = parser.parse_known_args()

# Initialize Organization
if args.organization == "None":
    args.organization = None
    args.desired_nodes = None
    args.time_to_live = None

# Initialize SlideRule Client
sliderule.init(args.domain, verbose=args.verbose, loglevel=args.loglvl, organization=args.organization, desired_nodes=args.desired_nodes, time_to_live=args.time_to_live)

# GEDI / 3DEP
def gedi_3dep():
    resource = "GEDI04_A_2019123154305_O02202_03_T00174_02_002_02_V002.h5"
    region = [  {"lon": -108.3435200747503, "lat": 38.89102961045247},
                {"lon": -107.7677425431139, "lat": 38.90611184543033},
                {"lon": -107.7818591266989, "lat": 39.26613714985466},
                {"lon": -108.3605610678553, "lat": 39.25086131372244},
                {"lon": -108.3435200747503, "lat": 38.89102961045247}  ]
    parms = {
        "poly": region,
        "degrade_filter": True,
        "l2_quality_filter": True,
        "beams": 0,
        "samples": {"3dep": {"asset": "usgs3dep-1meter-dem"}}
    }
    gdf = gedi.gedi04ap(parms, resources=[resource], as_numpy_array=True)
    samples = [values[0] for values in gdf["3dep.value"] if (type(values) == numpy.ndarray and not math.isnan(values[0]))]
    return sum(samples)/len(samples) # average value

# ICESat-2 / ArcticDEM
def icesat2_arcticdem():
    resource = "ATL03_20190314093716_11600203_007_01.h5"
    region = sliderule.toregion("tests/data/dicksonfjord.geojson")
    parms = { "poly": region['poly'],
              "cnf": "atl03_high",
              "srt": 3,
              "ats": 20.0,
              "cnt": 10,
              "len": 40.0,
              "res": 20.0,
              "maxi": 1,
              "samples": {"mosaic": {"asset": "arcticdem-mosaic"}} }
    gdf = icesat2.atl06p(parms, resources=[resource])
    return gdf["mosaic.value"].describe()["mean"]

# ICESat-2 / ATL06p
def icesat2_atl06p():
    resource = "ATL03_20181019065445_03150111_007_01.h5"
    parms = { "cnf": "atl03_high",
                "srt": 3,
                "ats": 20.0,
                "cnt": 10,
                "len": 40.0,
                "res": 20.0,
                "maxi": 1 }
    gdf = icesat2.atl06p(parms, resources=[resource])
    return gdf["h_mean"].mean()

# ICESat-2 / PhoREAL
def icesat2_phoreal():
    parms = {
        "poly": sliderule.toregion('tests/data/grandmesa.geojson')['poly'],
        "t0": '2019-11-14T00:00:00Z',
        "t1": '2019-11-15T00:00:00Z',
        "srt": icesat2.SRT_LAND,
        "len": 100,
        "res": 100,
        "pass_invalid": True,
        "atl08_class": ["atl08_ground", "atl08_canopy", "atl08_top_of_canopy"],
        "phoreal": {"binsize": 1.0, "geoloc": "center"}
    }
    gdf = sliderule.run("atl03x", parms)
    return gdf["h_mean_canopy"].mean()
#
# Main
#
print(f'GEDI / 3DEP = {gedi_3dep()}')
print(f'ICESat-2 / ArcticDEM = {icesat2_arcticdem()}')
print(f'ICESat-2 / ATL06p = {icesat2_atl06p()}')
print(f'ICESat-2 / PhoREAL = {icesat2_phoreal()}')


from sliderule import sliderule, earthdata
import sys

earthdata.set_max_resources(2000)

geojson = sys.argv[1]
region = sliderule.toregion(geojson)
granules = earthdata.cmr(short_name="GEDI02_A", polygon=region["poly"])

for granule in granules:
    print(f'{__file__[:__file__.rfind("/")]}/lpdaac_transfer.sh {granule}')



"""
sliderule
=========

A C++/Lua framework for on-demand science data processing
"""
from .sliderule import *
import sliderule.version

# get semantic version from setuptools-scm
__version__ = sliderule.version.version
